from cython cimport boundscheck, wraparound
import numpy as np
cimport numpy as np 

# ctypedef Int long long int
ctypedef np.int64_t Int
ctypedef np.float64_t Float


# The following fused typedefs are provided for
# cases when RHS is lowered to 32 bit, but their
# performance is terrible. Recompile instead for 32bit RHSes 
# ctypedef fused Float:
#     np.float32_t
#     np.float64_t

@boundscheck(False)
@wraparound(False)
def RHSAssemblyNative(np.ndarray[Float,ndim=2] T, np.ndarray[Float,ndim=2] t, 
    Int elem, Int nvar, Int nodeperelem,
    np.ndarray[unsigned long,ndim=2, mode='c'] elements):

    cdef long nelem = elements.shape[0]

    RHSAssemblyNative_(&T[0,0],&t[0,0],elem,nvar,nodeperelem,nelem,&elements[0,0])




cdef inline void RHSAssemblyNative_(Float *T, const Float *t, 
    Int elem, Int nvar, Int nodeperelem, long nelem, const unsigned long *elements):

    cdef Int i, T_idx, iterator

    # INTERNAL TRACTION FORCE ASSEMBLY
    for i in range(nodeperelem):
        T_idx = elements[elem*nodeperelem+i]*nvar
        for iterator in range(nvar):
            T[T_idx+iterator] += t[i*nvar+iterator]

    # For the record the following is the numpy equivalent version of above
    # for iterator in range(0,nvar):
        # T[mesh.elements[elem,:]*nvar+iterator,0]+=t[iterator::nvar,0]



from .VariationalPrinciple import *
from .DisplacementFormulation import *
from .DisplacementPotentialFormulation import *
from .LaplacianFormulation import *
from .ExplicitPenaltyContactFormulation import *
from .CoupleStressFormulation import *
from .FlexoelectricFormulation import *


import numpy as np
from Florence.FunctionSpace.OneDimensional import Line as OneD
from Florence.MeshGeneration.NodeArrangement import NodeArrangementQuad


def Lagrange(C,zeta,eta,arrange=1):
    """Computes higher order Lagrangian bases with equally spaced points
    """

    # Allocate
    Bases = np.zeros(((C+2)**2,1))
    Neta = np.zeros((C+2,1));   Nzeta = np.zeros((C+2,1))
    # Compute each from one-dimensional bases
    Nzeta[:,0] = OneD.Lagrange(C,zeta)[0]
    Neta[:,0] =  OneD.Lagrange(C,eta)[0]
    # Ternsorial product
    if arrange==1:
        node_arranger = NodeArrangementQuad(C)[2]
        Bases = np.dot(Neta,Nzeta.T).flatten()
        Bases = Bases[node_arranger]
        Bases = Bases[:,None]
    elif arrange==0:
        Bases[:,0] = np.dot(Neta,Nzeta.T).flatten()

    return Bases

def GradLagrange(C,zeta,eta,arrange=1):
    """Computes gradient of higher order Lagrangian bases with equally spaced points
    """

    # Allocate
    gBases = np.zeros(((C+2)**2,2))
    Nzeta = np.zeros((C+2,1));  Neta = np.zeros((C+2,1))
    gNzeta = np.zeros((C+2,1)); gNeta = np.zeros((C+2,1))
    # Compute each from one-dimensional bases
    Nzeta[:,0] = OneD.Lagrange(C,zeta)[0]
    Neta[:,0] = OneD.Lagrange(C,eta)[0]
    gNzeta[:,0] = OneD.Lagrange(C,zeta)[1]
    gNeta[:,0] = OneD.Lagrange(C,eta)[1]
    # Ternsorial product
    if arrange==1:
        node_arranger = NodeArrangementQuad(C)[2]
        # g0 = np.dot(gNeta,Nzeta.T).flatten()
        # g1 = np.dot(Neta,gNzeta.T).flatten()
        g0 = np.dot(Nzeta,gNeta.T).flatten()
        g1 = np.dot(gNzeta,Neta.T).flatten()
        gBases[:,0] = g0[node_arranger]
        gBases[:,1] = g1[node_arranger]
    elif arrange==0:
        gBases[:,0] = np.dot(gNeta,Nzeta.T).reshape((C+2)**2)
        gBases[:,1] = np.dot(Neta,gNzeta.T).reshape((C+2)**2)

    return gBases




import numpy as np 
import numpy.linalg as la
from Florence.FiniteElements.LocalAssembly.KinematicMeasures import *
from .PostProcess import *
from copy import copy




class ErrorNorms(PostProcess):

    def __init__(self,post_process=None,functors=None):
        
        # COPY THE WHOLE OBJECT / SHALLOW COPY
        self.ndim = post_process.ndim
        self.formulation = post_process.formulation
        self.domain_bases = post_process.domain_bases
        self.postdomain_bases = post_process.postdomain_bases
        self.boundary_bases = post_process.boundary_bases
        self.analysis_type = post_process.analysis_type
        self.analysis_nature = post_process.analysis_nature
        self.material_type = post_process.material_type
        self.functors = functors

    def SetExactSolution(self,functors):
        self.functors = functor

    def ExactSolution(self,func,*args,**kwargs):
        """Provide analytical/exact/closed-form solution through func"""
        return func(*args,**kwargs)


    def InterpolationBasedNormNonlinear(self,mesh,solution):
        """Compute error norms in convex multi-variable extended kinematic set {F,H,J,D0,d} 
            and the associated work conjugates - Based on work-conjugates of model 106"""

        if self.functors is None:
            raise ValueError("An exact solution functor/lambda function not provided")

        func = self.functors
        TotalDisp = solution

        Domain = self.domain_bases
        nodeperelem = mesh.elements.shape[1] 
        ndim = mesh.points.shape[1]

        if self.material is not None:
            mu1 = self.material.mu1
            mu2 = self.material.mu2
            lamb = self.material.lamb
            eps_1 = self.material.eps_1
            eps_2 = self.material.eps_2
        else:
            mu1, mu2, lamb = 1., 0.5, 1 
            eps_1, eps_2 = 4., 4.

        I = np.zeros((Domain.AllGauss.shape[0],ndim,ndim))
        for i in range(Domain.AllGauss.shape[0]):
            I[i,:,:] = np.eye(ndim,ndim)    

        # L2_normX = 0; L2_denormX = 0
        # L2_normE = 0.; L2_denormE = 0.
        L2_normx = 0.; L2_denormx = 0.
        L2_normF = 0.; L2_denormF = 0.
        L2_normH = 0.; L2_denormH = 0.
        L2_normJ = 0.; L2_denormJ = 0.
        L2_normPhi = 0.; L2_denormPhi = 0.
        L2_normD = 0.; L2_denormD = 0.
        L2_normd = 0.; L2_denormd = 0.
        L2_normSF = 0.; L2_denormSF = 0.
        L2_normSH = 0.; L2_denormSH = 0.
        L2_normSJ = 0.; L2_denormSJ = 0.
        L2_normSD = 0.; L2_denormSD = 0.
        L2_normSd = 0.; L2_denormSd = 0.

        for elem in range(mesh.nelem):

            # GET ELEMENTAL COORDINATES
            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            vpoints = mesh.points + TotalDisp[:,:,-1]
            EulerElemCoords = vpoints[mesh.elements[elem,:],:]

            # GET PHYSICAL ELEMENTAL GAUSS POINTS - USING FE BASES INTERPLOLATE COORDINATES AT GAUSS POINTS
            LagrangeElemGaussCoords = np.einsum('ij,ik',Domain.Bases,LagrangeElemCoords)
            EulerElemGaussCoords    = np.einsum('ij,ik',Domain.Bases,EulerElemCoords)

            # EVALAUTE STRAINS AT ALL GAUSS POINTS
            # ParentGradientX = np.einsum('ijk,jl->kil',Domain.Jm,LagrangeElemCoords)
            # ParentGradientx = np.einsum('ijk,jl->kil',Domain.Jm,EulerElemCoords)
            # MaterialGradient = np.einsum('ijk,kli->ijl',la.inv(ParentGradientX),Domain.Jm)
            # F = np.einsum('ij,kli->kjl', EulerElemCoords, MaterialGradient)
            # StrainTensors = KinematicMeasures(F, self.analysis_nature)


            # GEOMETRY ERROR
            GeomNodes = func.Exact_x(LagrangeElemCoords)
            ExactGeom = func.Exact_x(LagrangeElemGaussCoords)
            NumericalGeom = np.einsum('ij,ik->kj',GeomNodes,Domain.Bases)

            # DEFORMATION GRADIENT ERROR
            ExactEulerxNodes = GeomNodes + LagrangeElemCoords
            ParentGradientX = np.einsum('ijk,jl->kil',Domain.Jm,LagrangeElemCoords)
            MaterialGradient = np.einsum('ijk,kli->ijl',la.inv(ParentGradientX),Domain.Jm)
            F = np.einsum('ij,kli->kjl', ExactEulerxNodes, MaterialGradient)
            F_exact = func.Exact_F(LagrangeElemGaussCoords)
 
            # JACOBIAN ERROR
            J = np.linalg.det(F)
            J_exact = np.linalg.det(F_exact)

            # COFACTOR ERROR
            H = np.einsum('i,ijk->ijk',J,np.einsum('ikj',np.linalg.inv(F)))
            H_exact = np.einsum('i,ijk->ijk',J_exact,np.einsum('ikj',np.linalg.inv(F_exact)))

            # ELECTRIC POTENTIAL
            PotenialNodes = func.Exact_Phi(LagrangeElemCoords)
            ExactPotential = func.Exact_Phi(LagrangeElemGaussCoords)
            NumericalPotential = np.einsum('i,ij',PotenialNodes,Domain.Bases)

            # ELECTRIC FIELD ERROR
            ENodes = func.Exact_E(LagrangeElemCoords)
            E_exact = func.Exact_E(LagrangeElemGaussCoords)
            E = np.einsum('ij,ik->kj',ENodes,Domain.Bases)

            # ELECTRIC DISPLACEMENT ERROR
            C = np.einsum('ikj,ikl->ijl',F,F)
            C_exact = np.einsum('ikj,ikl->ijl',F,F)
            D = np.einsum('ijk,ik->ij',np.linalg.inv(eps_1*I+eps_2*C),E)
            D_exact = np.einsum('ijk,ik->ij',np.linalg.inv(eps_1*I+eps_2*C_exact),E_exact)

            # d = FD0 ERROR
            d = np.einsum('ijk,ik->ij',F,D)
            d_exact = np.einsum('ijk,ik->ij',F_exact,D_exact)


            # KINETICS - WORK CONJUGATES
            SF = 2.*mu1*F
            SF_exact = 2.*mu1*F_exact

            SH = 2.*mu2*H
            SH_exact = 2.*mu2*H_exact

            SJ = -2.*(mu1+2*mu2)*1./J + lamb*(J-1) - 1/eps_1/J**2*np.einsum('ij,ij->i',d,d)
            SJ_exact = -2.*(mu1+2*mu2)*1./J_exact + lamb*(J_exact-1) - 1/eps_1/J_exact**2*np.einsum('ij,ij->i',d_exact,d_exact)

            SD = 1./eps_1*D
            SD_exact = 1./eps_1*D_exact

            Sd = 1./eps_1*d
            Sd_exact = 1./eps_1*d_exact



            for counter in range(0,Domain.AllGauss.shape[0]):
            # for counter in range(1):    

                # L2_normx += (ExactGeom - NumericalGeom)**2*Domain.AllGauss[counter,0]
                # L2_denormx += (ExactGeom)**2*Domain.AllGauss[counter,0]

                # L2_normF += (F_exact - F)**2*Domain.AllGauss[counter,0]
                # L2_denormF += (F_exact)**2*Domain.AllGauss[counter,0]

                # L2_normH += (H_exact - H)**2*Domain.AllGauss[counter,0]
                # L2_denormH += (H_exact)**2*Domain.AllGauss[counter,0]

                # L2_normJ += (J_exact - J)**2*Domain.AllGauss[counter,0]
                # L2_denormJ += (J_exact)**2*Domain.AllGauss[counter,0]

                # L2_normPhi += (ExactPotential - NumericalPotential)**2*Domain.AllGauss[counter,0]
                # L2_denormPhi += (ExactPotential)**2*Domain.AllGauss[counter,0]

                # L2_normD += (D_exact - D)**2*Domain.AllGauss[counter,0]
                # L2_denormD += (D_exact)**2*Domain.AllGauss[counter,0]

                # L2_normd += (d_exact - d)**2*Domain.AllGauss[counter,0]
                # L2_denormd += (d_exact)**2*Domain.AllGauss[counter,0]

                # L2_normSF += (SF_exact - SF)**2*Domain.AllGauss[counter,0]
                # L2_denormSF += (SF_exact)**2*Domain.AllGauss[counter,0]

                # L2_normSH += (SH_exact - SH)**2*Domain.AllGauss[counter,0]
                # L2_denormSH += (SH_exact)**2*Domain.AllGauss[counter,0]

                # L2_normSJ += (SJ_exact - SJ)**2*Domain.AllGauss[counter,0]
                # L2_denormSJ += (SJ_exact)**2*Domain.AllGauss[counter,0]

                # L2_normSD += (SD_exact - SD)**2*Domain.AllGauss[counter,0]
                # L2_denormSD += SD_exact**2*Domain.AllGauss[counter,0]

                # L2_normSd += (Sd_exact - Sd)**2*Domain.AllGauss[counter,0]
                # L2_denormSd += Sd_exact**2*Domain.AllGauss[counter,0]

                ###########################################################

                L2_normx += (ExactGeom[counter,:] - NumericalGeom[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormx += (ExactGeom[counter,:])**2*Domain.AllGauss[counter,0]

                L2_normF += (F_exact[counter,:,:] - F[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormF += (F_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normH += (H_exact[counter,:,:] - H[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormH += (H_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normJ += (J_exact[counter] - J[counter])**2*Domain.AllGauss[counter,0]
                L2_denormJ += (J_exact[counter])**2*Domain.AllGauss[counter,0]

                L2_normPhi += (ExactPotential[counter] - NumericalPotential[counter])**2*Domain.AllGauss[counter,0]
                L2_denormPhi += (ExactPotential[counter])**2*Domain.AllGauss[counter,0]

                L2_normD += (D_exact[counter,:] - D[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormD += (D_exact[counter,:])**2*Domain.AllGauss[counter,0]

                L2_normd += (d_exact[counter,:] - d[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormd += (d_exact[counter,:])**2*Domain.AllGauss[counter,0]

                L2_normSF += (SF_exact[counter,:,:] - SF[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormSF += (SF_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normSH += (SH_exact[counter,:,:] - SH[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormSH += (SH_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normSJ += (SJ_exact[counter] - SJ[counter])**2*Domain.AllGauss[counter,0]
                L2_denormSJ += (SJ_exact[counter])**2*Domain.AllGauss[counter,0]

                L2_normSD += (SD_exact[counter,:] - SD[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormSD += SD_exact[counter,:]**2*Domain.AllGauss[counter,0]

                L2_normSd += (Sd_exact[counter,:] - Sd[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormSd += Sd_exact[counter,:]**2*Domain.AllGauss[counter,0]

                ###########################################################

            # L2_normSH = np.einsum('ijk,i',(SH_exact - SH)**2,Domain.AllGauss[:,0])
            # L2_denormSH = np.einsum('ijk,i',SH_exact**2,Domain.AllGauss[counter,0])

            # L2_normSJ = np.einsum('ijk,i',(SF_exact - SF)**2,Domain.AllGauss[:,0])
            # L2_denormSJ = np.einsum('ijk,i',SF_exact**2,Domain.AllGauss[:,0])

            # L2_normSD = np.einsum('ij,i',(SD_exact - SD)**2,Domain.AllGauss[:,0])
            # L2_denormSD = np.einsum('ij,i',SD_exact**2,Domain.AllGauss[:,0])

            # L2_normSd = np.sum(np.einsum('ij,k',(Sd_exact - Sd)**2,Domain.AllGauss[:,0]),axis=2)
            # L2_denormSd = np.sum(np.einsum('ij,k',Sd_exact**2,Domain.AllGauss[:,0]),axis=2)


        # L2NormX = np.sqrt(np.sum(L2_normX))/np.sqrt(np.sum(L2_denormX))
        # L2Normx = np.sqrt(np.sum(L2_normx))/np.sqrt(np.sum(L2_denormx))

        L2Normx = np.sum(L2_normx)/np.sum(L2_denormx)
        L2NormF = np.sum(L2_normF)/np.sum(L2_denormF)
        L2NormH = np.sum(L2_normH)/np.sum(L2_denormH)
        L2NormJ = np.sum(L2_normJ)/np.sum(L2_denormJ)
        L2NormPhi = np.sum(L2_normPhi)/np.sum(L2_denormPhi)
        L2NormD = np.sum(L2_normD)/np.sum(L2_denormD)
        L2Normd = np.sum(L2_normd)/np.sum(L2_denormd)
        L2NormSF = np.sum(L2_normSF)/np.sum(L2_denormSF)
        L2NormSH = np.sum(L2_normSH)/np.sum(L2_denormSH)
        L2NormSJ = np.sum(L2_normSJ)/np.sum(L2_denormSJ)
        L2NormSD = np.sum(L2_normSD)/np.sum(L2_denormSD)
        L2NormSd = np.sum(L2_normSd)/np.sum(L2_denormSd)


        return L2Normx, L2NormF, L2NormH, L2NormJ, L2NormPhi, L2NormD, L2Normd, L2NormSF, L2NormSH, L2NormSJ, L2NormSD, L2NormSd





    def InterpolationBasedNormNonlinearObjective(self,mesh,solution):
        """Compute error norms in convex multi-variable extended kinematic set {C,G,C,D0} 
            and the associated work conjugates - Based on work-conjugates of model 106"""

        if self.functors is None:
            raise ValueError("An exact solution functor/lambda function not provided")

        func = self.functors
        TotalDisp = solution

        Domain = self.domain_bases
        nodeperelem = mesh.elements.shape[1] 
        ndim = mesh.points.shape[1]

        if self.material is not None:
            mu1 = self.material.mu1
            mu2 = self.material.mu2
            lamb = self.material.lamb
            eps_1 = self.material.eps_1
            eps_2 = self.material.eps_2
        else:
            mu1, mu2, lamb = 1., 0.5, 1 
            eps_1, eps_2 = 4., 4.

        alpha, beta = 0.2, 0.2

        I = np.zeros((Domain.AllGauss.shape[0],ndim,ndim))
        for i in range(Domain.AllGauss.shape[0]):
            I[i,:,:] = np.eye(ndim,ndim)    


        L2_normx = 0.; L2_denormx = 0.
        L2_normC = 0.; L2_denormC = 0.
        L2_normG = 0.; L2_denormG = 0.
        L2_normdetC = 0.; L2_denormdetC = 0.
        L2_normPhi = 0.; L2_denormPhi = 0.
        L2_normD0 = 0.; L2_denormD0 = 0.
        L2_normSC = 0.; L2_denormSC = 0.
        L2_normSG = 0.; L2_denormSG = 0.
        L2_normSdetC = 0.; L2_denormSdetC = 0.
        L2_normSD0 = 0.; L2_denormSD0 = 0.

        for elem in range(mesh.nelem):

            # GET ELEMENTAL COORDINATES
            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            vpoints = mesh.points + TotalDisp[:,:,-1]
            EulerElemCoords = vpoints[mesh.elements[elem,:],:]

            # GET PHYSICAL ELEMENTAL GAUSS POINTS - USING FE BASES INTERPLOLATE COORDINATES AT GAUSS POINTS
            LagrangeElemGaussCoords = np.einsum('ij,ik',Domain.Bases,LagrangeElemCoords)
            EulerElemGaussCoords    = np.einsum('ij,ik',Domain.Bases,EulerElemCoords)

            # GEOMETRY ERROR
            GeomNodes = func.Exact_x(LagrangeElemCoords)
            ExactGeom = func.Exact_x(LagrangeElemGaussCoords)
            NumericalGeom = np.einsum('ij,ik->kj',GeomNodes,Domain.Bases)

            # DEFORMATION GRADIENT ERROR
            ExactEulerxNodes = GeomNodes + LagrangeElemCoords
            ParentGradientX = np.einsum('ijk,jl->kil',Domain.Jm,LagrangeElemCoords)
            MaterialGradient = np.einsum('ijk,kli->ijl',la.inv(ParentGradientX),Domain.Jm)
            F = np.einsum('ij,kli->kjl', ExactEulerxNodes, MaterialGradient)
            F_exact = func.Exact_F(LagrangeElemGaussCoords)
 
            # RIGHT CAUCHY GREEN
            C = np.einsum('ikj,ikl->ijl',F,F)
            C_exact = np.einsum('ikj,ikl->ijl',F_exact,F_exact)
            
            # JACOBIAN ERROR
            detC = np.abs(np.linalg.det(C))
            detC_exact = np.abs(np.linalg.det(C_exact))

            # COFACTOR ERROR
            G = np.einsum('i,ijk->ijk',detC,np.einsum('ikj',np.linalg.inv(C)))
            G_exact = np.einsum('i,ijk->ijk',detC_exact,np.einsum('ikj',np.linalg.inv(C_exact)))

            # ELECTRIC POTENTIAL ERROR
            PotenialNodes = func.Exact_Phi(LagrangeElemCoords)
            ExactPotential = func.Exact_Phi(LagrangeElemGaussCoords)
            NumericalPotential = np.einsum('i,ij',PotenialNodes,Domain.Bases)

            # ELECTRIC FIELD ERROR
            ENodes = func.Exact_E(LagrangeElemCoords)
            E_exact = func.Exact_E(LagrangeElemGaussCoords)
            E = np.einsum('ij,ik->kj',ENodes,Domain.Bases)

            # ELECTRIC DISPLACEMENT ERROR
            D = np.einsum('ijk,ik->ij',np.linalg.inv(eps_1*I+eps_2*C),E)
            D_exact = np.einsum('ijk,ik->ij',np.linalg.inv(eps_1*I+eps_2*C_exact),E_exact)

            # d = FD0 ERROR
            d = np.einsum('ijk,ik->ij',F,D)
            d_exact = np.einsum('ijk,ik->ij',F_exact,D_exact)


            # KINETICS - WORK CONJUGATES
            SC = 2.*mu1*I + 4.*mu1*alpha*np.einsum('ijk,ijk',C,I)*I
            SC_exact = 2.*mu1*I + 4.*mu1*alpha*np.einsum('ijk,ijk',C_exact,I)*I

            SG = 2.*mu2*I + 4.*mu2*beta*np.einsum('ijk,ijk',G,I)*I
            SG_exact = 2.*mu2*I + 4.*mu2*beta*np.einsum('ijk,ijk',G_exact,I)*I

            SdetC = -2.*(mu1+2*mu2)*1./detC + lamb*(1.-1./detC) - \
                1/2./eps_1/detC/np.sqrt(detC)*np.einsum('ij,ij->i',d,d)
            SdetC_exact = -2.*(mu1+2*mu2)*1./detC_exact + lamb*(1.-1./detC_exact) - \
                1/2./eps_1/detC_exact/np.sqrt(detC_exact)*np.einsum('ij,ij->i',d_exact,d_exact)

            SD = 1./eps_1*D + np.einsum('i,ij->ij',1./eps_2/np.sqrt(detC),np.einsum('ijk,ik->ij',C,D))
            SD_exact = 1./eps_1*D_exact + np.einsum('i,ij->ij',1./eps_2/np.sqrt(detC_exact),np.einsum('ijk,ik->ij',C_exact,D_exact))


            for counter in range(0,Domain.AllGauss.shape[0]):
            # for counter in range(1):
  
                ###########################################################

                L2_normx += (ExactGeom[counter,:] - NumericalGeom[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormx += (ExactGeom[counter,:])**2*Domain.AllGauss[counter,0]

                L2_normC += (C_exact[counter,:,:] - C[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormC += (C_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normG += (G_exact[counter,:,:] - G[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormG += (G_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normdetC += (detC_exact[counter] - detC[counter])**2*Domain.AllGauss[counter,0]
                L2_denormdetC += (detC_exact[counter])**2*Domain.AllGauss[counter,0]

                L2_normPhi += (ExactPotential[counter] - NumericalPotential[counter])**2*Domain.AllGauss[counter,0]
                L2_denormPhi += (ExactPotential[counter])**2*Domain.AllGauss[counter,0]

                L2_normD0 += (D_exact[counter,:] - D[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormD0 += (D_exact[counter,:])**2*Domain.AllGauss[counter,0]

                L2_normSC += (SC_exact[counter,:,:] - SC[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormSC += (SC_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normSG += (SG_exact[counter,:,:] - SG[counter,:,:])**2*Domain.AllGauss[counter,0]
                L2_denormSG += (SG_exact[counter,:,:])**2*Domain.AllGauss[counter,0]

                L2_normSdetC += (SdetC_exact[counter] - SdetC[counter])**2*Domain.AllGauss[counter,0]
                L2_denormSdetC += (SdetC_exact[counter])**2*Domain.AllGauss[counter,0]

                L2_normSD0 += (SD_exact[counter,:] - SD[counter,:])**2*Domain.AllGauss[counter,0]
                L2_denormSD0 += SD_exact[counter,:]**2*Domain.AllGauss[counter,0]

                ###########################################################

            # L2_normSd = np.sum(np.einsum('ij,k',(Sd_exact - Sd)**2,Domain.AllGauss[:,0]),axis=2)
            # L2_denormSd = np.sum(np.einsum('ij,k',Sd_exact**2,Domain.AllGauss[:,0]),axis=2)


        L2Normx = np.sum(L2_normx)/np.sum(L2_denormx)
        L2NormC = np.sum(L2_normC)/np.sum(L2_denormC)
        L2NormG = np.sum(L2_normG)/np.sum(L2_denormG)
        L2NormdetC = np.sum(L2_normdetC)/np.sum(L2_denormdetC)
        L2NormPhi = np.sum(L2_normPhi)/np.sum(L2_denormPhi)
        L2NormD0 = np.sum(L2_normD0)/np.sum(L2_denormD0)
        L2NormSC = np.sum(L2_normSC)/np.sum(L2_denormSC)
        L2NormSG = np.sum(L2_normSG)/np.sum(L2_denormSG)
        L2NormSdetC = np.sum(L2_normSdetC)/np.sum(L2_denormSdetC)
        L2NormSD0 = np.sum(L2_normSD0)/np.sum(L2_denormSD0)


        return L2Normx, L2NormC, L2NormG, L2NormdetC, L2NormPhi, L2NormD0, L2NormSC, L2NormSG, L2NormSdetC, L2NormSD0









## CHEAP NORM FOR CURVED MESH GENERATOR
def InterpolationBasedNorm(MainData,mesh,TotalDisp):

    if MainData.AssemblyParameters.FailedToConverge is True:
        MainData.L2NormX = np.NAN
        MainData.L2Normx = np.NAN
        MainData.DoF = mesh.points.shape[0]*MainData.nvar
        MainData.NELEM = mesh.elements.shape[0]

        return


    # func = lambda x,y : np.sin(x)*np.cos(y)
    # func = lambda x,y : x*np.sin(y)+y*np.cos(x)
    # func = lambda x,y : x**3+(y+1)**3
    # func = lambda x,y : (x+y)**11 ##
    func = lambda x,y : (x+y)**7 ##
    # func = lambda x,y : x**3*(y+1)**5

    nodeperelem = mesh.elements.shape[1]
    # print MainData.Domain.Bases[:,0].shape
    # print nodeperelem
    # print MainData.Domain.Bases[:,0].reshape(1,nodeperelem)
    # print MainData.Domain.Bases.shape
    # print MainData.Domain.Jm.shape

    # print MainData.Quadrature.points, "\n"

    # print MainData.Domain.Bases.shape, MainData.Domain.Jm.shape
    # exit()
    

    L2_normX = 0; L2_denormX = 0
    L2_normx = 0; L2_denormx = 0
    for elem in range(mesh.nelem):
        LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
        vpoints = mesh.points+TotalDisp[:,:,-1]
        EulerElemCoords = vpoints[mesh.elements[elem,:],:]
        # print LagrangeElemCoords.shape


        ParentGradientX = np.einsum('ijk,jl->kil',MainData.Domain.Jm,LagrangeElemCoords)
        ElementalSolGaussX = np.einsum('ij,ik',MainData.Domain.Bases,LagrangeElemCoords)

        AnalyticalSolNodesX = func(LagrangeElemCoords[:,0],LagrangeElemCoords[:,1])
        AnalyticalSolGaussX = func(ElementalSolGaussX[:,0],ElementalSolGaussX[:,1])
        NumericalSolGaussX = np.einsum('i,ij',AnalyticalSolNodesX,MainData.Domain.Bases)

        # Deformed
        ParentGradientx = np.einsum('ijk,jl->kil',MainData.Domain.Jm,EulerElemCoords)
        ElementalSolGaussx = np.einsum('ij,ik',MainData.Domain.Bases,EulerElemCoords)

        AnalyticalSolNodesx = func(EulerElemCoords[:,0],EulerElemCoords[:,1])
        AnalyticalSolGaussx = func(ElementalSolGaussx[:,0],ElementalSolGaussx[:,1])
        NumericalSolGaussx = np.einsum('i,ij',AnalyticalSolNodesx,MainData.Domain.Bases)

        # L2_norm = (AnalyticalSolGauss - NumericalSolGauss)

        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl',la.inv(ParentGradientX),MainData.Domain.Jm)


        # L2_norm += np.linalg.norm((AnalyticalSolGauss - NumericalSolGauss)**2)
        # L2_denorm += np.linalg.norm((AnalyticalSolGauss)**2)
        # print L2_norm, L2_denorm
        # print AnalyticalSolGauss -  NumericalSolGauss

        for counter in range(0,MainData.Domain.AllGauss.shape[0]):
            # L2_normX += np.linalg.norm((AnalyticalSolGaussX - NumericalSolGaussX)**2)*MainData.Domain.AllGauss[counter,0]
            # L2_denormX += np.linalg.norm((AnalyticalSolGaussX)**2)*MainData.Domain.AllGauss[counter,0]

            # L2_normx += np.linalg.norm((AnalyticalSolGaussx - NumericalSolGaussx)**2)*MainData.Domain.AllGauss[counter,0]
            # L2_denormx += np.linalg.norm((AnalyticalSolGaussx)**2)*MainData.Domain.AllGauss[counter,0]

            L2_normX += (AnalyticalSolGaussX - NumericalSolGaussX)**2*MainData.Domain.AllGauss[counter,0]
            L2_denormX += (AnalyticalSolGaussX)**2*MainData.Domain.AllGauss[counter,0]

            L2_normx += (AnalyticalSolGaussx - NumericalSolGaussx)**2*MainData.Domain.AllGauss[counter,0]
            L2_denormx += (AnalyticalSolGaussx)**2*MainData.Domain.AllGauss[counter,0]

            # L2_normX += (AnalyticalSolGaussX - NumericalSolGaussX)*MainData.Domain.AllGauss[counter,0]
            # L2_denormX += (AnalyticalSolGaussX)*MainData.Domain.AllGauss[counter,0]

            # L2_normx += (AnalyticalSolGaussx - NumericalSolGaussx)*MainData.Domain.AllGauss[counter,0]
            # L2_denormx += (AnalyticalSolGaussx)*MainData.Domain.AllGauss[counter,0]


    # L2NormX = np.sqrt(L2_normX)/np.sqrt(L2_denormX)
    # L2NormX = np.linalg.norm(L2NormX)

    # L2Normx = np.sqrt(L2_normx)/np.sqrt(L2_denormx)
    # L2Normx = np.linalg.norm(L2Normx)

    L2NormX = np.sqrt(np.sum(L2_normX))/np.sqrt(np.sum(L2_denormX))
    L2Normx = np.sqrt(np.sum(L2_normx))/np.sqrt(np.sum(L2_denormx))


    # L2NormX = L2_normX/L2_denormX
    # L2Normx = L2_normx/L2_denormx
    print(L2NormX, L2Normx)
    # print np.linalg.norm(L2NormX)

    MainData.L2NormX = L2NormX
    MainData.L2Normx = L2Normx
    MainData.DoF = mesh.points.shape[0]*MainData.nvar
    MainData.NELEM = mesh.elements.shape[0]






##
def ComputeErrorNorms(MainData,mesh,nmesh,AnalyticalSolution,Domain,Quadrature,MaterialArgs):

    # AT THE MOMENT THE ERROR NORMS ARE COMPUTED FOR LINEAR PROBLEMS ONLY
    if MainData.GeometryUpdate:
        raise ValueError('The error norms are computed for linear problems only.')

    # INITIATE/ALLOCATE
    C = MainData.C 
    ndim = MainData.ndim
    nvar = MainData.nvar
    elements = nmesh.elements
    points = nmesh.points
    nelem = elements.shape[0]
    nodeperelem = elements.shape[1]
    w = Quadrature.weights

    # TotalDisp & TotalPot ARE BOTH THIRD ORDER TENSOR (3RD DIMENSION FOR INCREMENTS) - TRANCATE THEM UNLESS REQUIRED
    TotalDisp = MainData.TotalDisp[:,:,-1]
    TotalPot  = MainData.TotalPot[:,:,-1]

    # # print TotalDisp
    # TotalDispa = np.zeros(TotalDisp.shape)
    # uxa = (points[:,1]*np.cos(points[:,0])); uxa=np.zeros(uxa.shape)
    # uya = (points[:,0]*np.sin(points[:,1]))
    # TotalDispa[:,0]=uxa
    # TotalDispa[:,1]=uya
    # # print TotalDispa
    # # print TotalDisp
    # print np.concatenate((TotalDispa,TotalDisp),axis=1)
    # # print points


    # ALLOCATE
    B = np.zeros((Domain.Bases.shape[0]*nvar,MaterialArgs.H_VoigtSize))
    E_nom = 0; E_denom = 0; L2_nom = 0; L2_denom = 0
    # LOOP OVER ELEMENTS
    for elem in range(0,nelem):
        xycoord = points[elements[elem,:],:]    
        # GET THE NUMERICAL SOLUTION WITHIN THE ELEMENT (AT NODES)
        ElementalSol = np.zeros((nodeperelem,nvar))
        ElementalSol[:,:ndim] = TotalDisp[elements[elem,:],:]
        ElementalSol[:,ndim]  = TotalPot[elements[elem,:],:].reshape(nodeperelem)
        # GET THE ANALYTICAL SOLUTION WITHIN THE ELEMENT (AT NODES)
        AnalyticalSolution.Args.node = xycoord
        AnalyticalSol = AnalyticalSolution().Get(AnalyticalSolution.Args)

        # AnalyticalSol[:,0] = ElementalSol[:,0]
        # print np.concatenate((AnalyticalSol[:,:2], ElementalSol[:,:2]),axis=1)
        # print

        # print points
        # print points[elements[elem,:],:]
        # print AnalyticalSol

        # ALLOCATE
        nvarBasis = np.zeros((Domain.Bases.shape[0],nvar))
        ElementalSolGauss  = np.zeros((Domain.AllGauss.shape[0],nvar)); AnalyticalSolGauss  = np.copy(ElementalSolGauss)
        dElementalSolGauss = np.zeros((Domain.AllGauss.shape[0],nvar)); dAnalyticalSolGauss = np.copy(ElementalSolGauss)
        # LOOP OVER GAUSS POINTS
        for counter in range(0,Domain.AllGauss.shape[0]):
            # GET THE NUMERICAL SOLUTION WITHIN THE ELEMENT (AT QUADRATURE POINTS)
            ElementalSolGauss[counter,:] = np.dot(Domain.Bases[:,counter].reshape(1,nodeperelem),ElementalSol)
            # GET THE ANALYTICAL SOLUTION WITHIN THE ELEMENT (AT QUADRATURE POINTS)
            AnalyticalSolution.Args.node = np.dot(Domain.Bases[:,counter],xycoord)
            AnalyticalSolGauss[counter,:] = AnalyticalSolution().Get(AnalyticalSolution.Args)

            # print AnalyticalSolGauss, ElementalSolGauss[:,0]
            # AnalyticalSolGauss[:,0] = ElementalSolGauss[:,0] # REMOVE
            # AnalyticalSolGauss[:,1] = ElementalSolGauss[:,1] # REMOVE
            # print np.concatenate((AnalyticalSolGauss[:,:2], ElementalSolGauss[:,:2]),axis=1)#, AnalyticalSolution.Args.node
            # print 
            # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
            Jm = Domain.Jm[:,:,counter]
            # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
            ParentGradientX=np.dot(Jm,xycoord) #
            # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
            MaterialGradient = np.dot(la.inv(ParentGradientX),Jm)
            # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
            # F = np.dot(EulerELemCoords.T,MaterialGradient.T)
            F = np.eye(ndim,ndim)
            # COMPUTE REMAINING KINEMATIC MEASURES
            StrainTensors = KinematicMeasures(F).Compute()
            # UPDATE/NO-UPDATE GEOMETRY
            if MainData.GeometryUpdate:
                # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
                ParentGradientx=np.dot(Jm,EulerELemCoords)
                # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
                SpatialGradient = np.dot(la.inv(ParentGradientx),Jm).T
            else:
                SpatialGradient = MaterialGradient.T

            # SPATIAL ELECTRIC FIELD
            ElectricFieldx = - np.dot(SpatialGradient.T,ElementalSol[:,ndim])
            # COMPUTE SPATIAL ELECTRIC DISPLACEMENT
            ElectricDisplacementx = MainData.ElectricDisplacementx(MaterialArgs,StrainTensors,ElectricFieldx)
            # COMPUTE CAUCHY STRESS TENSOR
            CauchyStressTensor = MainData.CauchyStress(MaterialArgs,StrainTensors,ElectricFieldx)
            # COMPUTE THE HESSIAN AT THIS GAUSS POINT
            H_Voigt = MainData.Hessian(MaterialArgs,ndim,StrainTensors,ElectricFieldx)
            # COMPUTE THE TANGENT STIFFNESS MATRIX
            BDB_1, t = MainData().ConstitutiveStiffnessIntegrand(Domain,B,MaterialGradient,nvar,SpatialGradient,
                ndim,CauchyStressTensor,ElectricDisplacementx,MaterialArgs,H_Voigt)

            # L2 NORM
            L2_nom   += np.linalg.norm((AnalyticalSolGauss - ElementalSolGauss)**2)*Domain.AllGauss[counter,0]
            L2_denom += np.linalg.norm((AnalyticalSolGauss)**2)*Domain.AllGauss[counter,0]

            # ENERGY NORM
            DiffSol = (AnalyticalSol - ElementalSol).reshape(ElementalSol.shape[0]*ElementalSol.shape[1],1)
            E_nom   += np.linalg.norm(DiffSol**2)*Domain.AllGauss[counter,0]
            E_denom += np.linalg.norm(AnalyticalSol.reshape(AnalyticalSol.shape[0]*AnalyticalSol.shape[1],1)**2)*Domain.AllGauss[counter,0]

    L2Norm = np.sqrt(L2_nom)/np.sqrt(L2_denom)
    EnergyNorm = np.sqrt(E_nom)/np.sqrt(E_denom)
    print(L2Norm, EnergyNorm)
    return L2Norm, EnergyNorm


import numpy as np
from .VariationalPrinciple import VariationalPrinciple
from Florence import QuadratureRule, FunctionSpace

from Florence.FiniteElements.LocalAssembly.KinematicMeasures import *
from Florence.FiniteElements.LocalAssembly._KinematicMeasures_ import _KinematicMeasures_
from Florence.Tensor import issymetric
from Florence.LegendreTransform import LegendreTransform

__all__ = ["ExplicitPenaltyContactFormulation"]

class ExplicitPenaltyContactFormulation(VariationalPrinciple):

    def __init__(self, mesh, plane_normal, distance, kappa, variables_order=(1,),
        quadrature_rules=None, quadrature_type=None, function_spaces=None, compute_post_quadrature=True,
        equally_spaced_bases=False, is_rigid_surface=True, contact_gap_tolerance=1e-6):
        """This is node-to-surface contact formulation based on penalty method"""

        self.plane_normal = np.array(plane_normal).flatten()
        self.distance = float(distance)
        self.kappa = float(kappa)
        self.contact_gap_tolerance = float(contact_gap_tolerance)
        self.is_rigid_surface = is_rigid_surface
        if self.is_rigid_surface:
            return

        if mesh.element_type != "tet" and mesh.element_type != "tri" and \
            mesh.element_type != "quad" and mesh.element_type != "hex":
            raise NotImplementedError( type(self).__name__, "has not been implemented for", mesh.element_type, "elements")

        if isinstance(variables_order,int):
            self.variables_order = (self.variables_order,)
        self.variables_order = variables_order

        super(ExplicitContactFormulation, self).__init__(mesh,variables_order=self.variables_order,
            quadrature_type=quadrature_type,quadrature_rules=quadrature_rules,function_spaces=function_spaces,
            compute_post_quadrature=compute_post_quadrature)

        self.fields = "mechanics"
        self.nvar = 0

        C = mesh.InferPolynomialDegree() - 1
        mesh.InferBoundaryElementType()

        if quadrature_rules == None and self.quadrature_rules == None:

            # OPTION FOR QUADRATURE TECHNIQUE FOR TRIS AND TETS
            optimal_quadrature = 3
            # is_flattened = True
            is_flattened = False

            if mesh.element_type == "tri" or mesh.element_type == "tet":
                norder = 2*C
                # TAKE CARE OF C=0 CASE
                if norder == 0:
                    norder = 1

                norder_post = 2*(C+1)
            else:
                norder = C+2
                norder_post = 2*(C+2)

            # GET QUADRATURE
            quadrature = QuadratureRule(optimal=optimal_quadrature, norder=norder, mesh_type=mesh.element_type, is_flattened=is_flattened)
            if self.compute_post_quadrature:
                # COMPUTE INTERPOLATION FUNCTIONS AT ALL INTEGRATION POINTS FOR POST-PROCESSING
                post_quadrature = QuadratureRule(optimal=optimal_quadrature, norder=norder_post, mesh_type=mesh.element_type)
            else:
                post_quadrature = None

            # BOUNDARY QUADRATURE
            bquadrature = QuadratureRule(optimal=optimal_quadrature, norder=C+2, mesh_type=mesh.boundary_element_type, is_flattened=is_flattened)

            self.quadrature_rules = (quadrature,post_quadrature,bquadrature)
        else:
            self.quadrature_rules = quadrature_rules

        if function_spaces == None and self.function_spaces == None:

            # CREATE FUNCTIONAL SPACES
            function_space = FunctionSpace(mesh, quadrature, p=C+1, equally_spaced=equally_spaced_bases, use_optimal_quadrature=is_flattened)
            if compute_post_quadrature:
                post_function_space = FunctionSpace(mesh, post_quadrature, p=C+1, equally_spaced=equally_spaced_bases)
            else:
                post_function_space = None

            # CREATE BOUNDARY FUNCTIONAL SPACES
            bfunction_space = FunctionSpace(mesh.CreateDummyLowerDimensionalMesh(),
                bquadrature, p=C+1, equally_spaced=equally_spaced_bases, use_optimal_quadrature=is_flattened)

            self.function_spaces = (function_space,post_function_space,bfunction_space)
        else:
            self.function_spaces = function_spaces


        # local_size = function_space.Bases.shape[0]*self.nvar
        local_size = self.function_spaces[0].Bases.shape[0]*self.nvar
        self.local_rows = np.repeat(np.arange(0,local_size),local_size,axis=0)
        self.local_columns = np.tile(np.arange(0,local_size),local_size)
        self.local_size = local_size

        # FOR MASS
        local_size_m = self.function_spaces[0].Bases.shape[0]*self.nvar
        self.local_rows_mass = np.repeat(np.arange(0,local_size_m),local_size_m,axis=0)
        self.local_columns_mass = np.tile(np.arange(0,local_size_m),local_size_m)
        self.local_size_m = local_size_m


    def GetElementalMatrices(self, elem, function_space, mesh, material, fem_solver, Eulerx, Eulerp):
        pass

    def GetElementalMatricesInVectorForm(self, elem, function_space, mesh, material, fem_solver, Eulerx, Eulerp):
        pass

    def GetLocalStiffness(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        pass

    def __GetLocalStiffness__(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        pass

    def ConstitutiveStiffnessIntegrand(self, B, SpatialGradient, ElectricDisplacementx,
            H_Voigt, analysis_nature="nonlinear", has_prestress=True):
        pass

    def GetLocalTraction(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get traction vector of the system"""
        pass

    def __GetLocalTraction__(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get traction vector of the system"""
        pass

    def TractionIntegrand(self, B, SpatialGradient, ElectricDisplacementx,
        CauchyStressTensor, analysis_nature="nonlinear", has_prestress=True):
        pass

    def GetLocalResidual(self):
        pass



    def AssembleTractions(self, mesh, material, Eulerx):
        """Assembly of contact forces is provided within the class itself"""

        # RIGID WALL LOCATED AT
        L = self.distance
        normal = self.plane_normal
        k = self.kappa
        ndim = mesh.points.shape[1]

        if ndim == 2:
            boundary_surface = mesh.edges
        else:
            boundary_surface = mesh.faces
        surfNodes_no, surfNodes_idx, surfNodes_inv = np.unique(boundary_surface, return_index=True, return_inverse=True)
        surfNodes = Eulerx[surfNodes_no,:]
        gNx = surfNodes.dot(normal) + L

        T_contact = np.zeros((mesh.points.shape[0]*ndim,1))
        contactNodes_idx = gNx < self.contact_gap_tolerance
        contactNodes = surfNodes[contactNodes_idx,:]

        # IF NO CONTACT
        if contactNodes.shape[0] == 0:
            return T_contact

        contactNodes_global_idx = surfNodes_no[contactNodes_idx].astype(np.int64)
        normal_gap = gNx[contactNodes_idx]

        # LOOP VERSION
        # for node in range(contactNodes.shape[0]):
        #     t_local = k*normal_gap[node]*normal
        #     T_contact[contactNodes_global_idx[node]*ndim:(contactNodes_global_idx[node]+1)*ndim,0] += t_local

        # VECTORISED VERSION
        T_contact = T_contact.reshape(mesh.points.shape[0],ndim)
        t_local = k*np.outer(normal_gap,normal)
        T_contact[contactNodes_global_idx,:] = t_local
        T_contact = T_contact.ravel()

        return T_contact



from __future__ import print_function, division
import numpy as np
import imp

# Pr = imp.load_source('Problem_Arch','BoundaryElements/Problem_Rectangle/ProblemDataBEM_1.py')
# Pr = imp.load_source('Problem_Arch','BoundaryElements/Problem_Arch/ProblemDataBEM_2.py')
# Pr = imp.load_source('Problem_Arch','BoundaryElements/Problem_Rectangle_LM/ProblemDataBEM_LM_1.py')
# Pr = imp.load_source('Problem_Arch','BoundaryElements/Problem_Arch_LM/ProblemDataBEM_LM_2.py')

def WritePlotBEM2D(sol,total_sol,POT,FLUX1,FLUX2,LHS2LHS,LHS2RHS,mesh,printt=1,plot=1,plotopt=1,write=0):

    # Separate potential and flux
    pot = sol[np.array(LHS2LHS,dtype=int)]
    dpot = sol[np.array(LHS2RHS,dtype=int)]

    if write==1:
        np.savetxt('Results.txt',total_sol)
        print('The BE solution at boundary point is written in text file "Results.txt" - both potential and flux')
        np.savetxt('POT.txt',total_sol)
        print('The potential values at internal points are written in text file "POT.txt"')
        np.savetxt('FLUX1.txt',total_sol)
        print('The X-direction flux values at internal points are written in text file "FLUX1.txt"')
        np.savetxt('FLUX2.txt',total_sol)
        print('The X-direction flux values at internal points are written in text file "FLUX2.txt"')

    if plot==1:
        # Call the plot function from the problem data file
        Pr.PlotFunc(mesh,POT,FLUX1,FLUX2,plotopt)

    if printt ==1:
        print(POT)
        # print(FLUX1)
        # print(FLUX2)


"""VTK output functions.

Use the XML VTK format for unstructured meshes (.vtu)

See here for a guide:  http://www.vtk.org/pdf/file-formats.pdf

Luke Olson 20090309
http://www.cs.uiuc.edu/homes/lukeo
"""

__docformat__ = "restructuredtext en"

__all__ = ['write_vtu','write_basic_mesh']

import xml.dom.minidom
import numpy

def write_vtu(Verts, Cells, pdata=None, pvdata=None, cdata=None, cvdata=None, fname='output.vtu'):
    """
    Write a .vtu file in xml format

    Parameters
    ----------
    fname : {string}
        file to be written, e.g. 'mymesh.vtu'
    Verts : {array}
        Ndof x 3 (if 2, then expanded by 0)
        list of (x,y,z) point coordinates
    Cells : {dictionary}
        Dictionary of with the keys
    pdata : {array}
        Ndof x Nfields array of scalar values for the vertices
    pvdata : {array}
        Nfields*3 x Ndof array of vector values for the vertices
    cdata : {dictionary}
        scalar valued cell data
    cvdata : {dictionary}
        vector valued cell data

    Returns
    -------
     writes a .vtu file for use in Paraview

    Notes
    -----
    - Poly data not supported
    - Non-Poly data is stored in Numpy array: Ncell x vtk_cell_info
    - Each I1 must be >=3
    - pdata = Ndof x Nfields
    - pvdata = 3*Ndof x Nfields
    - cdata,cvdata = list of dictionaries in the form of Cells


    =====  ========================= ============= ===
    keys   type                      n points      dim
    =====  ========================= ============= ===
       1   VTK_VERTEX:               1 point        2d
       2   VTK_POLY_VERTEX:          n points       2d
       3   VTK_LINE:                 2 points       2d
       4   VTK_POLY_LINE:            n+1 points     2d
       5   VTK_TRIANGLE:             3 points       2d
       6   VTK_TRIANGLE_STRIP:       n+2 points     2d
       7   VTK_POLYGON:              n points       2d
       8   VTK_PIXEL:                4 points       2d
       9   VTK_QUAD:                 4 points       2d
       10  VTK_TETRA:                4 points       3d
       11  VTK_VOXEL:                8 points       3d
       12  VTK_HEXAHEDRON:           8 points       3d
       13  VTK_WEDGE:                6 points       3d
       14  VTK_PYRAMID:              5 points       3d
       24  VTK_QUADRATIC_TETRA       10 points      3d
    =====  ========================= ============= ===

    Examples
    --------
    >>> import numpy
    >>> Verts = numpy.array([[0.0,0.0],
    ...                      [1.0,0.0],
    ...                      [2.0,0.0],
    ...                      [0.0,1.0],
    ...                      [1.0,1.0],
    ...                      [2.0,1.0],
    ...                      [0.0,2.0],
    ...                      [1.0,2.0],
    ...                      [2.0,2.0],
    ...                      [0.0,3.0],
    ...                      [1.0,3.0],
    ...                      [2.0,3.0]])
    >>> E2V = numpy.array([[0,4,3],
    ...                    [0,1,4],
    ...                    [1,5,4],
    ...                    [1,2,5],
    ...                    [3,7,6],
    ...                    [3,4,7],
    ...                    [4,8,7],
    ...                    [4,5,8],
    ...                    [6,10,9],
    ...                    [6,7,10],
    ...                    [7,11,10],
    ...                    [7,8,11]])
    >>> E2edge = numpy.array([[0,1]])
    >>> E2point = numpy.array([2,3,4,5])
    >>> Cells = {5:E2V,3:E2edge,1:E2point}
    >>> pdata=numpy.ones((12,2))
    >>> pvdata=numpy.ones((12*3,2))
    >>> cdata={5:numpy.ones((12,2)),3:numpy.ones((1,2)),1:numpy.ones((4,2))}
    >>> cvdata={5:numpy.ones((3*12,2)),3:numpy.ones((3*1,2)),1:numpy.ones((3*4,2))}
    >>> write_vtu(Verts=Verts, Cells=Cells, fname='test.vtu')

    See Also
    --------
    write_basic_mesh

    """
    # number of indices per cell for each cell type
    # vtk_cell_info = [-1, 1, None, 2, None, 3, None, None, 4, 4, 4, 8, 8, 6, 5]
    vtk_cell_info = [-1, 1, None, 2, None, 3, None, None, 4, 4, 4, 8, 8, 6, 5,
                        None, None, None, None, None, None, 3, 6, 8, 10, 20]

    # check fname
    if type(fname) is str:
        try:
            fname = open(fname,'w')
        except IOError:
            print(".vtu error (%s): %s" % (errno, strerror))
    else:
        raise ValueError('fname is assumed to be a string')

    # check Verts
    # get dimension and verify that it's 3d data
    Ndof,dim = Verts.shape
    if dim==2:
        # always use 3d coordinates (x,y) -> (x,y,0)
        Verts = numpy.hstack((Verts,numpy.zeros((Ndof,1))))

    # check Cells
    # keys must ve valid (integer and not "None" in vtk_cell_info)
    # Cell data can't be empty for a non empty key
    for key in Cells:
        if ((type(key) != int) or (key not in range(1,26))):
            raise ValueError('cell array must have positive integer keys in [1,25]')
        if (vtk_cell_info[key] == None) and (Cells[key] != None):
            # Poly data
            raise NotImplementedError('Poly Data not implemented yet')
        if Cells[key] is None:
            raise ValueError('cell array cannot be empty for key %d'%(key))
        if numpy.ndim(Cells[key])!=2:
            Cells[key] = Cells[key].reshape((Cells[key].size,1))
        if vtk_cell_info[key] != Cells[key].shape[1]:
            raise ValueError('cell array has %d columns, expected %d' % (offset, vtk_cell_info[key]) )

    # check pdata
    # must be Ndof x n_pdata
    n_pdata = 0
    if pdata is not None:
        if numpy.ndim(pdata)>1:
            n_pdata=pdata.shape[1]
        else:
            n_pdata = 1
            pdata = pdata.reshape((pdata.size,1))
        if pdata.shape[0] != Ndof:
            raise ValueError('pdata array should be of length %d (it is now %d)'%(Ndof,pdata.shape[0]))

    # check pvdata
    # must be 3*Ndof x n_pvdata
    n_pvdata = 0
    if pvdata is not None:
        if numpy.ndim(pvdata)>1:
            n_pvdata = pvdata.shape[1]
        else:
            n_pvdata = 1
            pvdata = pvdata.reshape((pvdata.size,1))
        if pvdata.shape[0] != 3*Ndof:
            raise ValueError('pvdata array should be of size %d (or multiples) (it is now %d)'%(Ndof*3,pvdata.shape[0]))

    # check cdata
    # must be NCells x n_cdata for each key
    n_cdata = 0
    if cdata !=None:
        for key in Cells:   # all valid now
            if numpy.ndim(cdata[key])>1:
                if n_cdata==0:
                    n_cdata=cdata[key].shape[1]
                elif n_cdata!=cdata[key].shape[1]:
                    raise ValueError('cdata dimension problem')
            else:
                n_cdata=1
                cdata[key] = cdata[key].reshape((cdata[key].size,1))
            if cdata[key].shape[0]!=Cells[key].shape[0]:
                raise ValueError('size mismatch with cdata %d and Cells %d'%(cdata[key].shape[0],Cells[key].shape[0]))
            if cdata[key] == None:
                raise ValueError('cdata array cannot be empty for key %d'%(key))

    # check cvdata
    # must be NCells*3 x n_cdata for each key
    n_cvdata = 0
    if cvdata is not None:
        for key in Cells:   # all valid now
            if numpy.ndim(cvdata[key])>1:
                if n_cvdata==0:
                    n_cvdata=cvdata[key].shape[1]
                elif n_cvdata!=cvdata[key].shape[1]:
                    raise ValueError('cvdata dimension problem')
            else:
                n_cvdata=1
                cvdata[key] = cvdata[key].reshape((cvdata[key].size,1))
            if cvdata[key].shape[0]!=3*Cells[key].shape[0]:
                raise ValueError('size mismatch with cvdata and Cells')
            if cvdata[key] is None:
                raise ValueError('cvdata array cannot be empty for key %d'%(key))

    Ncells = 0
    idx_min = 1

    cell_ind    = []
    cell_offset = [] #= numpy.zeros((Ncells,1),dtype=uint8) # offsets are zero indexed
    cell_type   = [] #= numpy.zeros((Ncells,1),dtype=uint8)

    cdata_all = None
    cvdata_all = None
    for key in Cells:
            # non-Poly data
            sz = Cells[key].shape[0]
            offset = Cells[key].shape[1]

            Ncells += sz
            cell_ind    = numpy.hstack((cell_ind,Cells[key].ravel()))
            cell_offset = numpy.hstack((cell_offset,offset*numpy.ones((sz,),dtype='uint8')))
            cell_type   = numpy.hstack((cell_type,key*numpy.ones((sz,),dtype='uint8')))

            if cdata != None:
                if cdata_all==None:
                    cdata_all=cdata[key]
                else:
                    cdata_all = numpy.vstack((cdata_all,cdata[key]))

            if cvdata != None:
                if cvdata_all==None:
                    cvdata_all=cvdata[key]
                else:
                    cvdata_all = numpy.vstack((cvdata_all,cvdata[key]))


    # doc element
    doc = xml.dom.minidom.Document()

    # vtk element
    root = doc.createElementNS('VTK', 'VTKFile')
    d = {'type':'UnstructuredGrid', 'version':'0.1', 'byte_order':'LittleEndian'}
    set_attributes(d,root)

    # unstructured element
    grid = doc.createElementNS('VTK', 'UnstructuredGrid')

    # piece element
    piece = doc.createElementNS('VTK', 'Piece')
    d = {'NumberOfPoints':str(Ndof),'NumberOfCells':str(Ncells)}
    set_attributes(d,piece)

    ## POINTS
    # points element
    points = doc.createElementNS('VTK', 'Points')
    # data element
    points_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type':'Float32', 'Name':'vertices', 'NumberOfComponents':'3', 'format':'ascii'}
    set_attributes(d,points_data)
    # string for data element
    points_data_str = doc.createTextNode(a2s(Verts))

    ## CELLS
    # points element
    cells = doc.createElementNS('VTK', 'Cells')
    # data element
    cells_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type':'Int32', 'Name':'connectivity', 'format':'ascii'}
    # d = {'type':'UInt8', 'Name':'connectivity', 'format':'ascii'}
    set_attributes(d,cells_data)
    # string for data element
    cells_data_str = doc.createTextNode(a2s(cell_ind))
    # offset data element
    cells_offset_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type':'Int32', 'Name':'offsets', 'format':'ascii'}
    # d = {'type':'UInt8', 'Name':'offsets', 'format':'ascii'}
    set_attributes(d,cells_offset_data)
    # string for data element
    cells_offset_data_str = doc.createTextNode(a2s(cell_offset.cumsum()))
    # offset data element
    cells_type_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type':'UInt8', 'Name':'types', 'format':'ascii'}
    set_attributes(d,cells_type_data)
    # string for data element
    cells_type_data_str = doc.createTextNode(a2s(cell_type))

    ## POINT DATA
    pointdata = doc.createElementNS('VTK', 'PointData')
    # pdata
    pdata_obj=[]
    pdata_str=[]
    for i in range(0,n_pdata):
        pdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        # d = {'type':'Float32', 'Name':'pdata %d'%(i), 'NumberOfComponents':'1', 'format':'ascii'}
        d = {'type':'Float32', 'Name':'Uz %d'%(i), 'NumberOfComponents':'1', 'format':'ascii'}
        set_attributes(d,pdata_obj[i])
        pdata_str.append(doc.createTextNode(a2s(pdata[:,i])))
    # pvdata
    pvdata_obj=[]
    pvdata_str=[]
    for i in range(0,n_pvdata):
        pvdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        d = {'type':'Float32', 'Name':'pvdata %d'%(i), 'NumberOfComponents':'3', 'format':'ascii'}
        set_attributes(d,pvdata_obj[i])
        pvdata_str.append(doc.createTextNode(a2s(pvdata[:,i])))

    ## CELL DATA
    celldata = doc.createElementNS('VTK', 'CellData')
    # cdata
    cdata_obj=[]
    cdata_str=[]
    for i in range(0,n_cdata):
        cdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        d = {'type':'Float32', 'Name':'cdata %d'%(i), 'NumberOfComponents':'1', 'format':'ascii'}
        set_attributes(d,cdata_obj[i])
        cdata_str.append(doc.createTextNode(a2s(cdata_all[:,i])))
    # cvdata
    cvdata_obj=[]
    cvdata_str=[]
    for i in range(0,n_cvdata):
        cvdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        d = {'type':'Float32', 'Name':'cvdata %d'%(i), 'NumberOfComponents':'3', 'format':'ascii'}
        set_attributes(d,cvdata_obj[i])
        cvdata_str.append(doc.createTextNode(a2s(cvdata_all[:,i])))

    doc.appendChild(root)
    root.appendChild(grid)
    grid.appendChild(piece)

    piece.appendChild(points)
    points.appendChild(points_data)
    points_data.appendChild(points_data_str)

    piece.appendChild(cells)
    cells.appendChild(cells_data)
    cells.appendChild(cells_offset_data)
    cells.appendChild(cells_type_data)
    cells_data.appendChild(cells_data_str)
    cells_offset_data.appendChild(cells_offset_data_str)
    cells_type_data.appendChild(cells_type_data_str)

    piece.appendChild(pointdata)
    for i in range(0,n_pdata):
        pointdata.appendChild(pdata_obj[i])
        pdata_obj[i].appendChild(pdata_str[i])
    for i in range(0,n_pvdata):
        pointdata.appendChild(pvdata_obj[i])
        pvdata_obj[i].appendChild(pvdata_str[i])

    piece.appendChild(celldata)
    for i in range(0,n_cdata):
        celldata.appendChild(cdata_obj[i])
        cdata_obj[i].appendChild(cdata_str[i])
    for i in range(0,n_cvdata):
        celldata.appendChild(cvdata_obj[i])
        cvdata_obj[i].appendChild(cvdata_str[i])

    doc.writexml(fname, newl='\n')
    fname.close()


def write_basic_mesh(Verts, E2V=None, mesh_type='tri', \
        pdata=None, pvdata=None, \
        cdata=None, cvdata=None, fname='output.vtu'):
    """
    Write mesh file for basic types of elements

    Parameters
    ----------
    fname : {string}
        file to be written, e.g. 'mymesh.vtu'
    Verts : {array}
        coordinate array (N x D)
    E2V : {array}
        element index array (Nel x Nelnodes)
    mesh_type : {string}
        type of elements: tri, quad, tet, hex (all 3d)
    pdata : {array}
        scalar data on vertices (N x Nfields)
    pvdata : {array}
        vector data on vertices (3*Nfields x N)
    cdata : {array}
        scalar data on cells (Nfields x Nel)
    cvdata : {array}
        vector data on cells (3*Nfields x Nel)

    Returns
    -------
    writes a .vtu file for use in Paraview

    Notes
    -----
    The difference between write_basic_mesh and write_vtu is that write_vtu is
    more general and requires dictionaries of cell information.
    write_basic_mesh calls write_vtu

    Examples
    --------
    >>> import numpy
    >>> Verts = numpy.array([[0.0,0.0],
    ...                      [1.0,0.0],
    ...                      [2.0,0.0],
    ...                      [0.0,1.0],
    ...                      [1.0,1.0],
    ...                      [2.0,1.0],
    ...                      [0.0,2.0],
    ...                      [1.0,2.0],
    ...                      [2.0,2.0],
    ...                      [0.0,3.0],
    ...                      [1.0,3.0],
    ...                      [2.0,3.0]])
    >>> E2V = numpy.array([[0,4,3],
    ...                    [0,1,4],
    ...                    [1,5,4],
    ...                    [1,2,5],
    ...                    [3,7,6],
    ...                    [3,4,7],
    ...                    [4,8,7],
    ...                    [4,5,8],
    ...                    [6,10,9],
    ...                    [6,7,10],
    ...                    [7,11,10],
    ...                    [7,8,11]])
    >>> pdata=numpy.ones((12,2))
    >>> pvdata=numpy.ones((12*3,2))
    >>> cdata=numpy.ones((12,2))
    >>> cvdata=numpy.ones((3*12,2))
    >>> write_basic_mesh(Verts, E2V=E2V, mesh_type='tri',pdata=pdata, pvdata=pvdata, cdata=cdata, cvdata=cvdata, fname='test.vtu')

    See Also
    --------
    write_vtu

    """
    if E2V is None:
        mesh_type='vertex'

    map_type_to_key = {'vertex':1, 'tri':5, 'quad':9, 'tet':10, 'hex':12}

    if mesh_type not in map_type_to_key:
        raise ValueError('unknown mesh_type=%s' % mesh_type)

    key = map_type_to_key[mesh_type]

    if mesh_type=='vertex':
        E2V = { key : numpy.arange(0,Verts.shape[0]).reshape((Verts.shape[0],1))}
    else:
        E2V = { key : E2V }

    if cdata != None:
        cdata = {key: cdata}

    if cvdata != None:
        cvdata = {key: cvdata}

    write_vtu(Verts=Verts, Cells=E2V, pdata=pdata, pvdata=pvdata, \
            cdata=cdata, cvdata=cvdata, fname=fname)


# ---------------------------------------
def set_attributes(d,elm):
    """
    helper function: Set attributes from dictionary of values
    """
    for key in d:
        elm.setAttribute(key,d[key])

def a2s(a):
    """
    helper function: Convert to string
    """
    str=''
    return str.join(['%g '%(v) for v in a.ravel()])



import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt, UnVoigt


class TranservselyIsotropicLinearElastic(Material):
    """A linear elastic transervely isotropic material model, with 4 material constants
        and 5 independent components in Hessian.

        Note that this assumes N = [0,0,1] as the direction of anisotropy
    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(TranservselyIsotropicLinearElastic, self).__init__(mtype,ndim,**kwargs)
        # MUST BE SET AFTER CALLING BASE __init__
        self.is_transversely_isotropic = True
        self.is_nonisotropic = True
        self.energy_type = "internal_energy"
        self.nature = "linear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        E = self.E
        E_A = self.E_A
        G_A = self.G_A
        v = self.nu

        H_Voigt = np.array([
                [ -(E*(- E*v**2 + E_A))/((v + 1)*(2*E*v**2 + E_A*v - E_A)),   -(E*v*(E_A + E*v))/((v + 1)*(2*E*v**2 + E_A*v - E_A)),      -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A),              0,              0,   0],
                [   -(E*v*(E_A + E*v))/((v + 1)*(2*E*v**2 + E_A*v - E_A)), -(E*(- E*v**2 + E_A))/((v + 1)*(2*E*v**2 + E_A*v - E_A)),      -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A),              0,              0,   0],
                [                     -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A),                     -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A), (E_A**2*(v - 1))/(2*E*v**2 + E_A*v - E_A),              0,              0,   0],
                [                                                      0,                                                      0,                                       0,                  E/(2*(v + 1)),  0,   0],
                [                                                      0,                                                      0,                                       0,                  0,              G_A, 0],
                [                                                      0,                                                      0,                                       0,                  0,              0, G_A]
            ])

        if self.ndim == 2:
            # CAREFUL WITH THIS SLICING AS SOME MATERIAL CONSTANTS WOULD BE REMOVED.
            # ESSENTIALLY IN PLANE STRAIN ANISOTROPY THE BEHAVIOUR OF MATERIAL
            # PERPENDICULAR TO THE PLANE IS LOST

            H_Voigt = H_Voigt[np.array([2,1,-1])[:,None],[2,1,-1]]


        self.H_VoigtSize = H_Voigt.shape[0]


        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        strain = StrainTensors['strain'][gcounter]
        strain_Voigt = Voigt(strain)

        E = self.E
        E_A = self.E_A
        G_A = self.G_A
        v = self.nu

        H_Voigt = np.array([
                [ -(E*(- E*v**2 + E_A))/((v + 1)*(2*E*v**2 + E_A*v - E_A)),   -(E*v*(E_A + E*v))/((v + 1)*(2*E*v**2 + E_A*v - E_A)),      -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A),              0,              0,   0],
                [   -(E*v*(E_A + E*v))/((v + 1)*(2*E*v**2 + E_A*v - E_A)), -(E*(- E*v**2 + E_A))/((v + 1)*(2*E*v**2 + E_A*v - E_A)),      -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A),              0,              0,   0],
                [                     -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A),                     -(E_A*E*v)/(2*E*v**2 + E_A*v - E_A), (E_A**2*(v - 1))/(2*E*v**2 + E_A*v - E_A),              0,              0,   0],
                [                                                      0,                                                      0,                                       0,                  E/(2*(v + 1)),  0,   0],
                [                                                      0,                                                      0,                                       0,                  0,              G_A, 0],
                [                                                      0,                                                      0,                                       0,                  0,              0, G_A]
            ])

        if self.ndim == 2:
            # CAREFUL WITH THIS SLICING AS SOME MATERIAL CONSTANTS WOULD BE REMOVED.
            # ESSENTIALLY IN PLANE STRAIN ANISOTROPY THE BEHAVIOUR OF MATERIAL
            # PERPENDICULAR TO THE PLANE IS LOST

            H_Voigt = H_Voigt[np.array([2,1,-1])[:,None],[2,1,-1]]

        stress = UnVoigt(np.dot(H_Voigt,strain_Voigt))

        return stress






from __future__ import division
import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material
from Florence.LegendreTransform import LegendreTransform


class IsotropicElectroMechanics_107(Material):
    """Electromechanical model in terms of internal energy 
        
            W(C,D0) = W_mn(C) + 1/2/eps_1 (D0*D0) + 1/2/eps_2/J (FD0*FD0) + W_reg(C,D0)
            W_mn(C) = u1*C:I+u2*G:I - 2*(u1+2*u2+6*mue)*lnJ + lamb/2*(J-1)**2 + mue*(C:I)**2
            W_reg(C,D_0) = 2/eps_e/mue (F:F)(FD0*FD0) + 1/mu/eps_e**2*(FD0*FD0)**2
    """
    
    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IsotropicElectroMechanics_107, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = self.ndim+1
        self.energy_type = "internal_energy"
        self.legendre_transform = LegendreTransform()
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"

        if self.ndim == 2:
            self.H_VoigtSize = 5
        elif self.ndim == 3:
            self.H_VoigtSize = 9

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx, elem=0):
        from Florence.MaterialLibrary.LLDispatch._IsotropicElectroMechanics_107_ import KineticMeasures
        return KineticMeasures(self,np.ascontiguousarray(F), ElectricFieldx)


    def Hessian(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        mue = self.mue
        lamb = self.lamb
        eps_1 = self.eps_1
        eps_2 = self.eps_2
        eps_e = self.eps_e

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        D  = ElectricDisplacementx.reshape(self.ndim,1)
        Dx = D.reshape(self.ndim)
        DD = np.dot(D.T,D)[0,0]
        D0D = np.dot(D,D.T)

        if self.ndim==2:
            trb = trace(b) + 1
        else:
            trb = trace(b)

        C_mech = 2.*mu2/J*(2*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b)) +\
            2.*(mu1+2.*mu2+6*mue)/J*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) ) + \
            lamb*(2.*J-1.)*einsum("ij,kl",I,I) - lamb*(J-1)*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) ) +\
            8.*mue/J*einsum('ij,kl',b,b)

        C_elect = 1./eps_2*(0.5*DD*(einsum('ik,jl',I,I) + einsum('il,jk',I,I) + einsum('ij,kl',I,I) ) - \
                einsum('ij,k,l',I,Dx,Dx) - einsum('i,j,kl',Dx,Dx,I))

        # REGULARISATION TERMS          
        C_reg = 8.*J/eps_e*( einsum('ij,k,l',b,Dx,Dx) + einsum('i,j,kl',Dx,Dx,b) ) + \
            8.*J**3/mue/eps_e**2*einsum('i,j,k,l',Dx,Dx,Dx,Dx)

        self.elasticity_tensor = C_mech + C_elect + C_reg

        self.coupling_tensor = 1./eps_2*(einsum('ik,j',I,Dx) + einsum('i,jk',Dx,I) - einsum('ij,k',I,Dx)) + \
            8*J/eps_e*einsum('ij,k',b,Dx) + 4*J/eps_e*trb*( einsum('ik,j',I,Dx) + einsum('i,jk',Dx,I) ) +\
            4.*J**3/mue/eps_e**2*DD* ( einsum('ik,j',I,Dx) + einsum('i,jk',Dx,I) ) + 8.*J**3/mue/eps_e**2*einsum('i,j,k',Dx,Dx,Dx)

        self.dielectric_tensor = J/eps_1*np.linalg.inv(b)  + 1./eps_2*I + \
            4.*J/eps_e*trb*I + \
            8.*J**3/mue/eps_e**2*(D0D + 0.5*DD*I)

        # TRANSFORM TENSORS TO THEIR ENTHALPY COUNTERPART
        E_Voigt, P_Voigt, C_Voigt = self.legendre_transform.InternalEnergyToEnthalpy(self.dielectric_tensor, 
            self.coupling_tensor, self.elasticity_tensor)
        
        # BUILD HESSIAN
        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*P_Voigt),axis=1)
        H2 = np.concatenate((factor*P_Voigt.T,E_Voigt),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)

        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        mue = self.mue
        lamb = self.lamb
        eps_2 = self.eps_2
        eps_e = self.eps_e

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        D = ElectricDisplacementx.reshape(self.ndim,1)
        DD = np.dot(D.T,D)[0,0]
        D0D = np.dot(D,D.T)

        if self.ndim==2:
            trb = trace(b) + 1
        else:
            trb = trace(b)

        sigma_mech = 2.0*mu1/J*b + \
            2.0*mu2/J*(trb*b - np.dot(b,b)) -\
            2.0*(mu1+2*mu2+6*mue)/J*I +\
            lamb*(J-1)*I +\
            4.*mue/J*trb*b
        sigma_electric = 1./eps_2*(D0D - 0.5*DD*I)
        # REGULARISATION TERMS
        sigma_reg = 4.*J/eps_e*(DD*b + trb*D0D) + \
            4.0*J**3/mue/eps_e**2*DD*D0D
        sigma = sigma_mech + sigma_electric + sigma_reg

        return sigma

    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        D = self.legendre_transform.GetElectricDisplacement(self, StrainTensors, ElectricFieldx, elem, gcounter)
        return D



    def Permittivity(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        mue = self.mue
        lamb = self.lamb
        eps_1 = self.eps_1
        eps_2 = self.eps_2
        eps_e = self.eps_e

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        D  = ElectricDisplacementx.reshape(self.ndim,1)
        Dx = D.reshape(self.ndim)
        DD = np.dot(D.T,D)[0,0]
        D0D = np.dot(D,D.T)

        if self.ndim==2:
            trb = trace(b) + 1
        else:
            trb = trace(b)

        self.dielectric_tensor = J/eps_1*np.linalg.inv(b)  + 1./eps_2*I + \
            4.*J/eps_e*trb*I + \
            8.*J**3/mue/eps_e**2*(D0D + 0.5*DD*I)

        # TRANSFORM TENSORS TO THEIR ENTHALPY COUNTERPART
        H_Voigt = -np.linalg.inv(self.dielectric_tensor)
        return H_Voigt


    def ElectrostaticMeasures(self,F,ElectricFieldx, elem=0):
        from Florence.MaterialLibrary.LLDispatch._IsotropicElectroMechanics_107_ import KineticMeasures
        D, _, H_Voigt = KineticMeasures(self,np.ascontiguousarray(F), ElectricFieldx)
        H_Voigt = np.ascontiguousarray(H_Voigt[:,-self.ndim:,-self.ndim:])
        return D, None, H_Voigt


from __future__ import division
import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material

class SteinmannModel(Material):
    """Steinmann's electromechanical model in terms of enthalpy

        W(C,E) = W_n(C) + c1*I:(E0 0 E0) + c2*C:(E0 0 E0) - eps_1/2*J*C**(-1):(E0 0 E0)
        W_n(C) = mu/2*(C:I-3) - mu*lnJ + lamb/2*(lnJ)**2

        Reference:
            D. K. Vu, P. Steinmann, and G. Possart, "Numerical modelling of non-linear electroelasticity",
            International Journal for Numerical Methods in Engineering, 70:685-704, (2007)   
    """
    
    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(SteinmannModel, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = self.ndim+1
        self.energy_type = "enthalpy"
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"

        if self.ndim == 2:
            self.H_VoigtSize = 5
        elif self.ndim == 3:
            self.H_VoigtSize = 9

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx, elem=0):
        from Florence.MaterialLibrary.LLDispatch._SteinmannModel_ import KineticMeasures
        return KineticMeasures(self, np.ascontiguousarray(F), ElectricFieldx)

    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        c1 = self.c1
        c2 = self.c2
        eps_1 = self.eps_1

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]


        E = 1.0*ElectricFieldx.reshape(self.ndim,1)
        Ex = E.reshape(E.shape[0])
        EE = np.dot(E,E.T)
        be = np.dot(b,ElectricFieldx).reshape(self.ndim)

        C_Voigt = lamb/J*einsum('ij,kl',I,I) - (lamb*np.log(J) - mu)/J*( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) + \
            eps_1*( einsum('ij,kl',I,EE) + einsum('ij,kl',EE,I) - einsum('ik,jl',EE,I) - einsum('il,jk',EE,I) - \
                einsum('ik,jl',I,EE) - einsum('il,jk',I,EE) ) + \
            eps_1*(np.dot(E.T,E)[0,0])*0.5*( einsum('ik,jl',I,I) + einsum('il,jk',I,I) - einsum('ij,kl',I,I) )

        C_Voigt = Voigt(C_Voigt,1)

        P_Voigt = eps_1*( einsum('ik,j',I,Ex) + einsum('jk,i',I,Ex) - einsum('ij,k',I,Ex)) +\
            2.0*c2/J*( einsum('ik,j',b,be) + einsum('i,jk',be,b) )
        
        P_Voigt = Voigt(P_Voigt,1)
            

        E_Voigt = -eps_1*I +  2.0*c1/J*b + 2.0*c2/J*np.dot(b,b)

        # Build the Hessian
        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*P_Voigt),axis=1)
        H2 = np.concatenate((factor*P_Voigt.T,E_Voigt),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)

        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        c1 = self.c1
        c2 = self.c2
        eps_1 = self.eps_1

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        E = ElectricFieldx.reshape(self.ndim,1)

        be = np.dot(b,E)

        stress = 1.0*mu/J*(b-I) + lamb/J*np.log(J)*I + \
            eps_1*(np.dot(E,E.T) - 0.5*np.dot(E.T,E)[0,0]*I) +\
            2.0*c2/J*np.dot(be,be.T)

        return stress


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        
        mu = self.mu
        lamb = self.lamb
        c1 = self.c1
        c2 = self.c2
        eps_1 = self.eps_1      

        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        E = ElectricFieldx.reshape(self.ndim,1)
        bb =  np.dot(b,b)
        be = np.dot(b,E)
        
        D = eps_1*E - 2.0*c1/J*be - 2.0*c2/J*np.dot(bb,E)

        return D























# import numpy as np
# from Florence.Tensor import trace
# #####################################################################################################
#                                 # Isotropic Steinmann Model
# #####################################################################################################


# class Steinmann(object):
#     """docstring for Steinmann"""
#     def __init__(self, ndim):
#         super(Steinmann, self).__init__()
#         self.ndim = ndim
#     def Get(self):
#         self.nvar = self.ndim+1
#         self.modelname = 'Steinmann'
#         return self.nvar, self.modelname

#     def Hessian(self,MaterialArgs,ndim,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

#         # Using Einstein summation (using numpy einsum call)
#         d = np.einsum

#         # Get material constants (5 in this case)
#         mu = MaterialArgs.mu
#         lamb = MaterialArgs.lamb
#         c1 = MaterialArgs.c1
#         c2 = MaterialArgs.c2
#         varepsilon_1 = MaterialArgs.eps_1

#         I = StrainTensors['I']
#         J = StrainTensors['J'][gcounter]
#         b = StrainTensors['b'][gcounter]

#         # Update Lame constants
#         mu2 = mu - lamb*(J-1.0)
#         lamb2 = lamb*(2.0*J-1.0) - mu

#         E = 1.0*ElectricFieldx
#         Ex = E.reshape(E.shape[0])
#         EE = np.dot(E,E.T)
#         be = np.dot(b,ElectricFieldx).reshape(ndim)

#         # Fourth order elasticity tensor
#         # C_Voigt = lamb2*AijBkl(I,I) +mu2*(AikBjl(I,I)+AilBjk(I,I)) +\
#         #   varepsilon_1*(AijBkl(I,EE) + AijBkl(EE,I) -AikBjl(EE,I)-AilBjk(EE,I)-AilBjk(I,EE)-AikBjl(I,EE) ) +\
#         #   varepsilon_1*(np.dot(E.T,E)[0,0])*(0.5*(AikBjl(I,I) + AilBjk(I,I))-0.5*AijBkl(I,I))
#         # C_Voigt=0.5*(C_Voigt+C_Voigt.T)

#         C_Voigt = Voigt(            
#             lamb2*d('ij,kl',I,I)+mu2*(d('ik,jl',I,I)+d('il,jk',I,I)) +\
#             varepsilon_1*(d('ij,kl',I,EE) + d('ij,kl',EE,I) - d('ik,jl',EE,I)-d('il,jk',EE,I)-d('il,jk',I,EE)-d('ik,jl',I,EE) ) +\
#             varepsilon_1*(np.dot(E.T,E)[0,0])*(0.5*(d('ik,jl',I,I) + d('il,jk',I,I))-0.5*d('ij,kl',I,I))
#             ,1
#             )

        
#         # Coupled Tensor (e - 3rd order)
#         # Note that the actual piezoelectric tensor is symmetric wrt to the last two indices
#         # Actual tensor (varepsilon_1 bit) is: e[k,i,j] += 1.0*varepsilon_1*(E[i]*delta[j,k] + E[j]*delta[i,k] - delta[i,j]*E[k]) 
#         # We need to make its Voigt_form symmetric with respect to (j,k) instead of (i,j) 
#         # e_voigt = 1.0*varepsilon_1*(AijUk(I,Ex)+AikUj(I,Ex)-UiAjk(Ex,I)).T +\
#         # (2.0*c2/J)*(AikUj(b,be)+AijUk(b,be)).T

#         e_voigt = Voigt(
#             1.0*varepsilon_1*(d('ij,k',I,Ex)+d('ik,j',I,Ex)-d('i,jk',Ex,I)) +\
#             (2.0*c2/J)*(d('ik,j',b,be)+d('ij,k',b,be))
#             ,1
#             )
            

#         # Dielectric Tensor (Permittivity - 2nd order)
#         Permittivity = -varepsilon_1*I +\
#         (2.0*c1/J)*b +\
#         (2.0*c2/J)*np.dot(b,b)

#         # Build the Hessian
#         factor = -1.
#         H1 = np.concatenate((C_Voigt,factor*e_voigt),axis=1)
#         H2 = np.concatenate((factor*e_voigt.T,Permittivity),axis=1)
#         H_Voigt = np.concatenate((H1,H2),axis=0)

#         MaterialArgs.H_VoigtSize = H_Voigt.shape[0]

#         return H_Voigt



#     def CauchyStress(self,MaterialArgs,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

#         c2 = MaterialArgs.c2

#         I = StrainTensors['I']
#         J = StrainTensors['J'][gcounter]
#         b = StrainTensors['b'][gcounter]
#         E = ElectricFieldx

#         mu = MaterialArgs.mu
#         lamb = MaterialArgs.lamb
#         varepsilon_1 = MaterialArgs.eps_1

#         be = np.dot(b,ElectricFieldx)

#         return 1.0*mu/J*b+(lamb*(J-1.0)-mu)*I + varepsilon_1*(np.dot(E,E.T)-0.5*np.dot(E.T,E)[0,0]*I) +\
#         (2.0*c2/J)*np.dot(be,be.T)


#     def ElectricDisplacementx(self,MaterialArgs,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        
#         c1 = MaterialArgs.c1
#         c2 = MaterialArgs.c2
#         varepsilon_1 = MaterialArgs.eps_1       

#         J = StrainTensors['J'][gcounter]
#         b = StrainTensors['b'][gcounter]
#         bb =  np.dot(b,b)
        
#         return varepsilon_1*ElectricFieldx -\
#         (2.0*c1/J)*np.dot(b,ElectricFieldx) -\
#         (2.0*c2/StrainTensors.J)*np.dot(bb,ElectricFieldx).reshape(StrainTensors.b.shape[0],1)




import numpy as np
cimport numpy as np
import cython

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

DTYPEI = np.int64
ctypedef np.int64_t DTYPEI_t

@cython.boundscheck(False)
@cython.wraparound(False)
def NPFROMFILE_Loop_Cython(np.ndarray[DTYPE_t, ndim=1] FileContent,DTYPEI_t nnode):
	# DETERMINE MULTIPLICITY OF EACH TYPE (i.e. EDGES, FACES, ELEMENTS)
	cdef DTYPEI_t cols1 = np.int64((FileContent[4*nnode+2+1] - 100) + 2)
	cdef DTYPEI_t cols2 = 0
	cdef DTYPEI_t counter_edge = 0
	cdef DTYPEI_t i 
	for i in range(4*nnode+2+1,FileContent.shape[0]):
		if FileContent[4*nnode+2+1+cols1*counter_edge] == FileContent[4*nnode+2+1]:
			counter_edge +=1
		else:
			cols2 = np.int64((FileContent[4*nnode+2+1+cols1*counter_edge] - 200)+2)
			break
	cdef DTYPEI_t counter_face = 0
	for i in range(4*nnode+2+1+cols1*counter_edge,FileContent.shape[0]):
		if 4*nnode+2+1+cols1*counter_edge+cols2*counter_face >= FileContent.shape[0]:
			break
		else:
			if FileContent[4*nnode+2+1+cols1*counter_edge+cols2*counter_face] == FileContent[4*nnode+2+1+cols1*counter_edge]:
				counter_face +=1
			else:
				break
	
	cdef np.ndarray nelse_type = np.zeros((2,2),dtype=np.int64); 
	nelse_type[0,0] = np.int64(FileContent[4*nnode+2+1]); nelse_type[1,0] = np.int64(FileContent[4*nnode+2+1+cols1*counter_edge])
	nelse_type[0,1] = counter_edge; nelse_type[1,1]=counter_face

	return nelse_type 


import numpy as np
from warnings import warn
from .Numeric import tovoigt, tovoigt3


__all__ = ['unique2d','in2d','intersect2d','in2d_unsorted','shuffle_along_axis',
'itemfreq','SecondTensor2Vector','Voigt','UnVoigt', 'remove_duplicates_2D','totuple']


#-------------------------------------------------------------------------#
# UTILITY FUNCTIONS

def unique2d(arr,axis=1,consider_sort=False,order=True,return_index=False,return_inverse=False):
    """Get unique values along an axis for a 2D array.
        see: http://stackoverflow.com/questions/16970982/find-unique-rows-in-numpy-array

        input:

            arr:
                2D array

            axis:
                Axis along which to take unique values, for instance unique
                rows (axis=1) or unique columns (axis=0). The axis ordering
                should not be confused with the usual numpy style axis argument,
                as here since unique values of a 1D type array is finally computed,
                numpy style axis hence becomes meaningless. Hence, in this context
                axis=0 implies finding unique rows of an array (lumping the rows) and
                axis=1 implies finding unique columns of an array (lumping the columns)

            consider_sort:
                Does permutation of the values in row/column matter. Two rows/columns
                can have the same elements but with different arrangements. If consider_sort
                is True then those rows/columns would be considered equal

            order:
                Similar to 1D unique in numpy, wherein the unique values are always sorted,
                if order is True, unique2d will also sort the values

            return_index:
                Similar to numpy unique. If order is True the indices would be sorted

            return_inverse:
                Similar to numpy unique

        returns:

            2D array of unique values
            If return_index is True also returns indices
            If return_inverse is True also returns the inverse array

            """

    if arr.ndim == 1:
        warn("1D is array is passed to 2D routine. Use the numpy.unique instead")
        arr = arr[:,None]

    if axis == 0:
        arr = np.copy(arr.T,order='C')

    if consider_sort is True:
        a = np.sort(arr,axis=1)
    else:
        a = arr
    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))

    if return_inverse is False:
        _, idx = np.unique(b, return_index=True)
    else:
        _, idx, inv = np.unique(b, return_index=True, return_inverse=True)

    if order is True:
        idx.sort()

    if return_index == False and return_inverse == False:
        return arr[idx]
    elif return_index == True and return_inverse == False:
        return arr[idx], idx
    elif return_index == False and return_inverse == True:
        return arr[idx], inv
    elif return_index == True and return_inverse == True:
        return arr[idx], idx, inv
    else:
        return arr[idx]



def in2d(arr1, arr2, axis=1, consider_sort=False):
    """Generalisation of numpy.in1d to 2D arrays

        NOTE: arr_1 and arr_2 should have the same dtype

        input:

            arr1:
                2D array

            arr2:
                2D array

            axis:
                Axis along which to np.in1d values, for instance along
                rows (axis=1) or along columns (axis=0). The axis ordering
                should not be confused with the usual numpy style axis argument,
                as here since np.in1d values of a 1D type array is finally computed,
                numpy style axis hence becomes meaningless. Hence, in this context
                axis=0 implies finding intersection rows of an array (lumping the rows) and
                axis=1 implies finding intersection columns of an array (lumping the columns)

            consider_sort:
                Does permutation of the values in row/column matter. Two rows/columns
                can have the same elements but with different arrangements. If consider_sort
                is True then those rows/columns would be considered equal

        returns:

            1D boolean array of the same length as arr1

            """

    assert arr1.dtype == arr2.dtype

    if axis == 0:
        arr1 = np.copy(arr1.T,order='C')
        arr2 = np.copy(arr2.T,order='C')

    if consider_sort is True:
        sorter_arr1 = np.argsort(arr1)
        arr1 = arr1[np.arange(arr1.shape[0])[:,None],sorter_arr1]
        sorter_arr2 = np.argsort(arr2)
        arr2 = arr2[np.arange(arr2.shape[0])[:,None],sorter_arr2]

    arr1_view = np.ascontiguousarray(arr1).view(np.dtype((np.void, arr1.dtype.itemsize * arr1.shape[1])))
    arr2_view = np.ascontiguousarray(arr2).view(np.dtype((np.void, arr2.dtype.itemsize * arr2.shape[1])))
    intersected = np.in1d(arr1_view, arr2_view)
    return intersected.view(np.bool).reshape(-1)


def intersect2d(arr1, arr2,axis=1, consider_sort=False):
    """Generalisation of numpy.intersect1d to 2D arrays

        NOTE: arr_1 and arr_2 should have the same dtype

        input:

            arr1:
                2D array

            arr2:
                2D array

            axis:
                Axis along which to take intersect1d values, for instance along
                rows (axis=1) or along columns (axis=0). The axis ordering
                should not be confused with the usual numpy style axis argument,
                as here since intersect1d values of a 1D type array is finally computed,
                numpy style axis hence becomes meaningless. Hence, in this context
                axis=0 implies finding intersection rows of an array (lumping the rows) and
                axis=1 implies finding intersection columns of an array (lumping the columns)

            consider_sort:
                Does permutation of the values in row/column matter. Two rows/columns
                can have the same elements but with different arrangements. If consider_sort
                is True then those rows/columns would be considered equal

        returns:

            2D array of intersection

            """

    assert arr1.dtype == arr2.dtype

    if axis == 0:
        arr1 = np.copy(arr1.T,order='C')
        arr2 = np.copy(arr2.T,order='C')

    if consider_sort is True:
        sorter_arr1 = np.argsort(arr1)
        arr1 = arr1[np.arange(arr1.shape[0])[:,None],sorter_arr1]
        sorter_arr2 = np.argsort(arr2)
        arr2 = arr2[np.arange(arr2.shape[0])[:,None],sorter_arr2]

    arr1_view = np.ascontiguousarray(arr1).view(np.dtype((np.void, arr1.dtype.itemsize * arr1.shape[1])))
    arr2_view = np.ascontiguousarray(arr2).view(np.dtype((np.void, arr2.dtype.itemsize * arr2.shape[1])))
    intersected = np.intersect1d(arr1_view, arr2_view)
    return intersected.view(arr1.dtype).reshape(-1, arr1.shape[1])



def in2d_unsorted(arr1, arr2, axis=1, consider_sort=False):
    """Find the elements in arr1 which are also in arr2 and sort them as they
        appear in arr2

        In principle this is similar to in2d but with two major differences:
        1. in2d is based on numpy.in1d which finds elements in arr1 which are
            also in arr2 but returns a non-sorted boolean array,
            for instance consider:

            >>> a=np.array([7, 3, 1, 5, 6, 2, 0, 4])
            >>> b=np.array([5, 1, 0, 4])
            >>> in1d(a,b)
            array([False, False,  True,  True, False, False,  True,  True], dtype=bool)

            whose nonzeros would give the indices as
            >>> _.nonzero()
            (array([2, 3, 6, 7]),)

            This is correct but the indices are sorted in the sense that the zeroth
            element in b (b[0]=5) is the 3rd element of a (a[3]=5) and the first element in
            b (b[1]=1) is the 2nd element in a (a[2]=1). This method preserves this ordering
            and returns

            (array([3, 2, 6, 7]),)

            instead

        2. This function returns indices instead of boolean array

        NOTE: arr_1 and arr_2 should have the same dtype

        input:

            arr1:
                2D array

            arr2:
                2D array

            axis:
                Axis along which to take unique values, for instance unique
                rows (axis=1) or unique columns (axis=0). The axis ordering
                should not be confused with the usual numpy style axis argument,
                as here since unique values of a 1D type array is finally computed,
                numpy style axis hence becomes meaningless. Hence, in this context
                axis=0 implies finding unique rows of an array (lumping the rows) and
                axis=1 implies finding unique columns of an array (lumping the columns)

            consider_sort:
                Does permutation of the values in row/column matter. Two rows/columns
                can have the same elements but with different arrangements. If consider_sort
                is True then those rows/columns would be considered equal

        returns:

            1D index array of elements common elements as they appear in arr2

            """

    assert arr1.dtype == arr2.dtype

    if axis == 0:
        arr1 = np.copy(arr1.T,order='C')
        arr2 = np.copy(arr2.T,order='C')

    if consider_sort is True:
        sorter_arr1 = np.argsort(arr1)
        arr1 = arr1[np.arange(arr1.shape[0])[:,None],sorter_arr1]
        sorter_arr2 = np.argsort(arr2)
        arr2 = arr2[np.arange(arr2.shape[0])[:,None],sorter_arr2]


    arr = np.vstack((arr1,arr2))
    _, inv = unique2d(arr, return_inverse=True)

    size1 = arr1.shape[0]
    size2 = arr2.shape[0]

    arr3 = inv[:size1]
    arr4 = inv[-size2:]

    sorter = np.argsort(arr3)
    idx = sorter[arr3.searchsorted(arr4, sorter=sorter)]

    return idx


def shuffle_along_axis(A,B,axis=1,consider_sort=False):
    """Given two equal (but shuffled) 2D arrays A and B, shuffles B along an specified
        axis (rows or columns) such that B==A

        input:
            A:
                2D array

            B:
                2D array to shuffle

            axis:
                Axis along which to take unique values, for instance unique
                rows (axis=1) or unique columns (axis=0). The axis ordering
                should not be confused with the usual numpy style axis argument,
                as here since unique values of a 1D type array is finally computed,
                numpy style axis hence becomes meaningless. Hence, in this context
                axis=0 implies finding unique rows of an array (lumping the rows) and
                axis=1 implies finding unique columns of an array (lumping the columns)

            consider_sort:
                Does permutation of the values in row/column matter. Two rows/columns
                can have the same elements but with different arrangements. If consider_sort
                is True then those rows/columns would be considered equal


        returns:
            A_to_B:
                A mapper such that A[A_to_B] == B
    """

    assert A.dtype == B.dtype
    assert A.shape == B.shape

    if axis == 0:
        A = np.copy(A.T,order='C')
        B = np.copy(B.T,order='C')

    if consider_sort is True:
        sorter_arr1 = np.argsort(A)
        A = A[np.arange(A.shape[0])[:,None],sorter_arr1]
        sorter_arr2 = np.argsort(B)
        B = B[np.arange(B.shape[0])[:,None],sorter_arr2]


    rowtype = np.dtype((np.void, A.dtype.itemsize * A.shape[1]))
    a = np.ascontiguousarray(A).view(rowtype).ravel()
    b = np.ascontiguousarray(B).view(rowtype).ravel()

    A_to_As = np.argsort(a)

    As_to_B = a[A_to_As].searchsorted(b)
    A_to_B = A_to_As[As_to_B]

    return A_to_B



def itemfreq(arr=None,un_arr=None,inv_arr=None,decimals=None):

    if (arr is None) and (un_arr is None):
        raise ValueError('No input array to work with. Either the array or its unique with inverse should be provided')
    if un_arr is None:
        if decimals is not None:
            un_arr, inv_arr = np.unique(np.round(arr,decimals=decimals),return_inverse=True)
        else:
            un_arr, inv_arr = np.unique(arr,return_inverse=True)

    if arr is not None:
        if len(arr.shape) > 1:
            dtype = type(arr[0,0])
        else:
            dtype = type(arr[0])
    else:
        if len(un_arr.shape) > 1:
            dtype = type(un_arr[0,0])
        else:
            dtype = type(un_arr[0])

    unf_arr = np.zeros((un_arr.shape[0],2),dtype=dtype)
    unf_arr[:,0] = un_arr
    unf_arr[:,1] = np.bincount(inv_arr)

    return unf_arr



def remove_duplicates_2D(A, decimals=10):
    """Removes duplicates from floating point 2D array (A) with rounding
    """

    assert isinstance(A,np.ndarray)
    assert (A.dtype == np.float64 or A.dtype == np.float32)

    from Florence.Tensor import makezero
    makezero(A)
    rounded_repoints = np.round(A,decimals=decimals)
    _, idx_repoints, inv_repoints = unique2d(rounded_repoints,order=False,
        consider_sort=False,return_index=True,return_inverse=True)
    A = A[idx_repoints,:]

    return A, idx_repoints, inv_repoints




def SecondTensor2Vector(A):

    # Check size of the matrix
    if A.shape[0]>3 or A.shape[0]==1 or A.shape[0]!=A.shape[1]:
        raise ValueError('Only square 2x2 and 3x3 matrices can be transformed to Voigt vector form')
    if A.shape[0]==3:
        # Matrix is symmetric
        vecA = np.array([
            A[0,0],A[1,1],A[2,2],A[0,1],A[0,2],A[1,2]
            ])
        # Check for symmetry
        if not np.allclose(A.T, A, rtol=1e-12,atol=1e-15):
            # Matrix is non-symmetric
            vecA = np.array([
                A[0,0],A[1,1],A[2,2],A[0,1],A[0,2],A[1,2],A[1,0],A[2,0],A[2,1]
                ])
    elif A.shape[0]==2:
        # Matrix is symmetric
        vecA = np.array([
            A[0,0],A[1,1],A[0,1]
            ])
        # Check for symmetry
        if not np.allclose(A.T, A, rtol=1e-12,atol=1e-15):
            # Matrix is non-symmetric
            vecA = np.array([
                A[0,0],A[1,1],A[0,1],A[1,0]
                ])

    return vecA


def Voigt(A,sym=1):
    """Given a 4th order tensor A, puts it in 6x6 format
        Given a 3rd order tensor A, puts it in 3x6 format (symmetric wrt the first two indices)
        Given a 2nd order tensor A, puts it in 1x6 format

        sym returns the symmetrised tensor (only for 3rd and 4th order). Switched on by default
        """

    if sym==0:
        return __voigt_unsym__(A)


    if A.ndim==4:
         VoigtA = tovoigt(A)
    elif A.ndim==3:
        VoigtA = tovoigt3(A)
    elif A.ndim==2:
        VoigtA = SecondTensor2Vector(A)
        # PURE PYTHON VERSION
        # e=A
        # if e.shape[0]==3:
        #     VoigtA = 0.5*np.array([
        #         [2.*e[0,0,0],2.*e[0,0,1],2.*e[0,0,2]],
        #         [2.*e[1,1,0],2.*e[1,1,1],2.*e[1,1,2]],
        #         [2.*e[2,2,0],2.*e[2,2,1],2.*e[2,2,2]],
        #         [e[0,1,0]+e[1,0,0],e[0,1,1]+e[1,0,1],e[0,1,2]+e[1,0,2]],
        #         [e[0,2,0]+e[2,0,0],e[0,2,1]+e[2,0,1],e[0,2,2]+e[2,0,2]],
        #         [e[1,2,0]+e[2,1,0],e[1,2,1]+e[2,1,1],e[1,2,2]+e[2,1,2]],
        #         ])
        # elif e.shape[0]==2:
        #     VoigtA = 0.5*np.array([
        #         [2.*e[0,0,0],2.*e[0,0,1]],
        #         [2.*e[1,1,0],2.*e[1,1,1]],
        #         [e[0,1,0]+e[1,0,0],e[0,1,1]+e[1,0,1]]
        #         ])

    return VoigtA


def __voigt_unsym__(A):

    if A.ndim==4:
        C=A
        if C.shape[0]==3:
            VoigtA = np.array([
                [C[0,0,0,0],C[0,0,1,1],C[0,0,2,2],C[0,0,0,1],C[0,0,0,2],C[0,0,1,2]],
                [C[1,1,0,0],C[1,1,1,1],C[1,1,2,2],C[1,1,0,1],C[1,1,0,2],C[1,1,1,2]],
                [C[2,2,0,0],C[2,2,1,1],C[2,2,2,2],C[2,2,0,1],C[2,2,0,2],C[2,2,1,2]],
                [C[0,1,0,0],C[0,1,1,1],C[0,1,2,2],C[0,1,0,1],C[0,1,0,2],C[0,1,1,2]],
                [C[0,2,0,0],C[0,2,1,1],C[0,2,2,2],C[0,2,0,1],C[0,2,0,2],C[0,2,1,2]],
                [C[1,2,0,0],C[1,2,1,1],C[1,2,2,2],C[1,2,0,1],C[1,2,0,2],C[1,2,1,2]]
                ])
        elif C.shape[0]==2:
            VoigtA = np.array([
                [C[0,0,0,0],C[0,0,1,1],C[0,0,0,1]],
                [C[1,1,0,0],C[1,1,1,1],C[1,1,0,1]],
                [C[0,1,0,0],C[0,1,1,1],C[0,1,0,1]]
                ])

    elif A.ndim==3:
        e = A
        if e.shape[0]==3:
            VoigtA = np.array([
                [e[0,0,0],e[0,0,1],e[0,0,2]],
                [e[1,1,0],e[1,1,1],e[1,1,2]],
                [e[2,2,0],e[2,2,1],e[2,2,2]],
                [e[0,1,0],e[0,1,1],e[0,1,2]],
                [e[0,2,0],e[0,2,1],e[0,2,2]],
                [e[1,2,0],e[1,2,1],e[1,2,2]],
                ])
        else:
            VoigtA = np.array([
                [e[0,0,0],e[0,0,1]],
                [e[1,1,0],e[1,1,1]],
                [e[0,1,0],e[0,1,1]]
                ])

    elif A.ndim==2:
        VoigtA = SecondTensor2Vector(A)

    else:
        VoigtA = np.array([])

    return VoigtA




def UnVoigt(v):
    A = []
    if v.ndim==2:
        if v.shape[1]>1:
            pass
            # DO IT LATER FOR shape>1
        elif v.shape[1]==1:
            # VECTORS TO SYMMETRIC 2ND ORDER TENSORS
            if v.shape[0]==6:
                A = np.array([
                    [v[0,0],v[3,0],v[4,0]],
                    [v[3,0],v[1,0],v[5,0]],
                    [v[4,0],v[5,0],v[2,0]]
                    ])
            elif v.shape[0]==3:
                A = np.array([
                    [v[0,0],v[2,0]],
                    [v[2,0],v[1,0]]
                    ])

    elif v.ndim==1:
        # VECTORS TO SYMMETRIC 2ND ORDER TENSORS
        if v.shape[0]==6:
            A = np.array([
                [v[0],v[3],v[4]],
                [v[3],v[1],v[5]],
                [v[4],v[5],v[2]]
                ])
        elif v.shape[0]==3:
            A = np.array([
                [v[0],v[2]],
                [v[2],v[1]]
                ])

    return A



def totuple(arr):
    """Converts numpy array to tuple"""
    return tuple(map(tuple, np.atleast_2d(arr)))














def IncrementallyLinearisedStress(Stress_k,H_Voigt_k,I,strain,Gradu):
    # IN PRINCIPLE WE NEED GRADU AND NOT STRAIN FOR V_strain
    V_strain =  Voigt(strain)[:,None]
                    # STRESS                        HESSSIAN 'I_W:GRADU'
    return np.dot(Stress_k,(I+strain)) + UnVoigt( np.dot(H_Voigt_k,V_strain) )










########################################################################################################
# Note that these matrices are all symmetrised
def AijBkl(A,B):

    A=1.0*A; B=1.0*B
    A00=A[0,0]; A11=A[1,1]; A22=A[2,2]; A01=A[0,1]; A02=A[0,2]; A12=A[1,2]; A10=A[1,0]; A20=A[2,0]; A21=A[2,1]
    B00=B[0,0]; B11=B[1,1]; B22=B[2,2]; B01=B[0,1]; B02=B[0,2]; B12=B[1,2]; B10=B[1,0]; B20=B[2,0]; B21=B[2,1]

    Tens = 1.0*np.array([
        [ A00*B00, A00*B11, A00*B22, A00*B01, A00*B02, A00*B12],
        [ A11*B00, A11*B11, A11*B22, A11*B01, A11*B02, A11*B12],
        [ A22*B00, A22*B11, A22*B22, A22*B01, A22*B02, A22*B12],
        [ A01*B00, A01*B11, A01*B22, A01*B01, A01*B02, A01*B12],
        [ A02*B00, A02*B11, A02*B22, A02*B01, A02*B02, A02*B12],
        [ A12*B00, A12*B11, A12*B22, A12*B01, A12*B02, A12*B12]
        ])


    return Tens


def AikBjl(A,B):

    A=1.0*A; B=1.0*B
    A00=A[0,0]; A11=A[1,1]; A22=A[2,2]; A01=A[0,1]; A02=A[0,2]; A12=A[1,2]; A10=A[1,0]; A20=A[2,0]; A21=A[2,1]
    B00=B[0,0]; B11=B[1,1]; B22=B[2,2]; B01=B[0,1]; B02=B[0,2]; B12=B[1,2]; B10=B[1,0]; B20=B[2,0]; B21=B[2,1]

    Tens = 1.0*np.array([
        [ A00*B00, A01*B01, A02*B02, A00*B01, A00*B02, A01*B02],
        [ A10*B10, A11*B11, A12*B12, A10*B11, A10*B12, A11*B12],
        [ A20*B20, A21*B21, A22*B22, A20*B21, A20*B22, A21*B22],
        [ A00*B10, A01*B11, A02*B12, A00*B11, A00*B12, A01*B12],
        [ A00*B20, A01*B21, A02*B22, A00*B21, A00*B22, A01*B22],
        [ A10*B20, A11*B21, A12*B22, A10*B21, A10*B22, A11*B22]
        ])


    return Tens


def AilBjk(A,B):

    A=1.0*A; B=1.0*B
    A00=A[0,0]; A11=A[1,1]; A22=A[2,2]; A01=A[0,1]; A02=A[0,2]; A12=A[1,2]; A10=A[1,0]; A20=A[2,0]; A21=A[2,1]
    B00=B[0,0]; B11=B[1,1]; B22=B[2,2]; B01=B[0,1]; B02=B[0,2]; B12=B[1,2]; B10=B[1,0]; B20=B[2,0]; B21=B[2,1]

    Tens = 1.0*np.array([
        [ A00*B00, A01*B01, A02*B02, A01*B00, A02*B00, A02*B01],
        [ A10*B10, A11*B11, A12*B12, A11*B10, A12*B10, A12*B11],
        [ A20*B20, A21*B21, A22*B22, A21*B20, A22*B20, A22*B21],
        [ A00*B10, A01*B11, A02*B12, A01*B10, A02*B10, A02*B11],
        [ A00*B20, A01*B21, A02*B22, A01*B20, A02*B20, A02*B21],
        [ A10*B20, A11*B21, A12*B22, A11*B20, A12*B20, A12*B21]
        ])


    return Tens


# A TENSOR AND A VECTOR
def AijUk(A,U):

    A=1.0*A; U=1.0*U
    A00=A[0,0]; A11=A[1,1]; A22=A[2,2]; A01=A[0,1]; A02=A[0,2]; A12=A[1,2]; A10=A[1,0]; A20=A[2,0]; A21=A[2,1]
    U0=U[0]; U1=U[1]; U2=U[2]

    Tens = 1.0*np.array([
        [ A00*U0, A01*U1, A02*U2, A00*U1, A00*U2, A01*U2],
        [ A10*U0, A11*U1, A12*U2, A10*U1, A10*U2, A11*U2],
        [ A20*U0, A21*U1, A22*U2, A20*U1, A20*U2, A21*U2]
        ])

    return Tens


def AikUj(A,U):

    A=1.0*A; U=1.0*U
    A00=A[0,0]; A11=A[1,1]; A22=A[2,2]; A01=A[0,1]; A02=A[0,2]; A12=A[1,2]; A10=A[1,0]; A20=A[2,0]; A21=A[2,1]
    U0=U[0]; U1=U[1]; U2=U[2]

    Tens = 1.0*np.array([
        [ A00*U0, A01*U1, A02*U2, A01*U0, A02*U0, A02*U1],
        [ A10*U0, A11*U1, A12*U2, A11*U0, A12*U0, A12*U1],
        [ A20*U0, A21*U1, A22*U2, A21*U0, A22*U0, A22*U1]
        ])


    return Tens


def UiAjk(U,A):

    A=1.0*A; U=1.0*U
    A00=A[0,0]; A11=A[1,1]; A22=A[2,2]; A01=A[0,1]; A02=A[0,2]; A12=A[1,2]; A10=A[1,0]; A20=A[2,0]; A21=A[2,1]
    U0=U[0]; U1=U[1]; U2=U[2]

    Tens = 1.0*np.array([
        [ A00*U0, A11*U0, A22*U0, A01*U0, A02*U0, A12*U0],
        [ A00*U1, A11*U1, A22*U1, A01*U1, A02*U1, A12*U1],
        [ A00*U2, A11*U2, A22*U2, A01*U2, A02*U2, A12*U2]
        ])


    return Tens









from __future__ import print_function
import gc, os, sys
import multiprocessing
from copy import deepcopy
from warnings import warn
from time import time
import numpy as np
from numpy.linalg import norm
import scipy as sp
from Florence.Utils import insensitive

from Florence.FiniteElements.Assembly import Assemble
from Florence.PostProcessing import *
from Florence.Solver import LinearSolver
from Florence.TimeIntegrators import StructuralDynamicIntegrators
from Florence import Mesh, FEMSolver


# class FEMSolverArcLength(FEMSolver):

#     def __init__(self):
#         pass


def StaticSolverArcLength(self, function_spaces, formulation, solver, K,
        NeumannForces, NodalForces, Residual,
        mesh, TotalDisp, Eulerx, Eulerp, material, boundary_condition):

    LoadIncrement = self.number_of_load_increments
    # LoadFactor = 1./LoadIncrement
    AppliedDirichletInc = np.zeros(boundary_condition.applied_dirichlet.shape[0],dtype=np.float64)
    # self.incremental_load_factor = 0.
    self.incremental_load_factor = 1./LoadIncrement
    self.accumulated_load_factor = 0.
    self.arc_length_scaling_factor = 1.0
    
    for Increment in range(LoadIncrement):

        # APPLY NEUMANN BOUNDARY CONDITIONS
        DeltaF = self.incremental_load_factor*NeumannForces
        NodalForces += DeltaF
        # OBRTAIN INCREMENTAL RESIDUAL - CONTRIBUTION FROM BOTH NEUMANN AND DIRICHLET
        Residual = -boundary_condition.ApplyDirichletGetReducedMatrices(K,Residual,
            boundary_condition.applied_dirichlet,LoadFactor=self.incremental_load_factor,only_residual=True)
        Residual -= DeltaF
        # GET THE INCREMENTAL DISPLACEMENT
        AppliedDirichletInc = self.incremental_load_factor*boundary_condition.applied_dirichlet

        t_increment = time()

        # LET NORM OF THE FIRST RESIDUAL BE THE NORM WITH RESPECT TO WHICH WE
        # HAVE TO CHECK THE CONVERGENCE OF NEWTON RAPHSON. TYPICALLY THIS IS 
        # NORM OF NODAL FORCES
        if Increment==0:
            self.NormForces = np.linalg.norm(Residual)
            # AVOID DIVISION BY ZERO
            if np.isclose(self.NormForces,0.0):
                self.NormForces = 1e-14

        self.norm_residual = np.linalg.norm(Residual)/self.NormForces

        Eulerx, Eulerp, K, Residual = NewtonRaphsonArchLength(self, function_spaces, formulation, solver, 
            Increment, K, NodalForces, Residual, mesh, Eulerx, Eulerp,
            material, boundary_condition, AppliedDirichletInc, NeumannForces, TotalDisp)

        # UPDATE DISPLACEMENTS FOR THE CURRENT LOAD INCREMENT
        TotalDisp[:,:formulation.ndim,Increment] = Eulerx - mesh.points
        if formulation.fields == "electro_mechanics":
            TotalDisp[:,-1,Increment] = Eulerp

        # PRINT LOG IF ASKED FOR
        if self.print_incremental_log:
            dmesh = Mesh()
            dmesh.points = TotalDisp[:,:formulation.ndim,Increment]
            dmesh_bounds = dmesh.Bounds
            if formulation.fields == "electro_mechanics":
                _bounds = np.zeros((2,formulation.nvar))
                _bounds[:,:formulation.ndim] = dmesh_bounds
                _bounds[:,-1] = [TotalDisp[:,-1,Increment].min(),TotalDisp[:,-1,Increment].max()]
                print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),_bounds)
            else:
                print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),dmesh_bounds)

        # SAVE INCREMENTAL SOLUTION IF ASKED FOR
        if self.save_incremental_solution:
            from scipy.io import savemat
            if self.incremental_solution_filename is not None:
                savemat(self.incremental_solution_filename+"_"+str(Increment),{'solution':TotalDisp[:,:,Increment]},do_compression=True)
            else:
                raise ValueError("No file name provided to save incremental solution")


        print('\nFinished Load increment', Increment, 'in', time()-t_increment, 'seconds')
        try:
            print('Norm of Residual is', 
                np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces), '\n')
        except RuntimeWarning:
            print("Invalid value encountered in norm of Newton-Raphson residual")

        # STORE THE INFORMATION IF NEWTON-RAPHSON FAILS
        if self.newton_raphson_failed_to_converge:
            solver.condA = np.NAN
            TotalDisp = TotalDisp[:,:,:Increment]
            self.number_of_load_increments = Increment
            break

        # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
        if self.break_at_increment != -1 and self.break_at_increment is not None:
            if self.break_at_increment == Increment:
                if self.break_at_increment < LoadIncrement - 1:
                    print("\nStopping at increment {} as specified\n\n".format(Increment))
                    TotalDisp = TotalDisp[:,:,:Increment]
                break


    return TotalDisp


def NewtonRaphsonArchLength(self, function_spaces, formulation, solver, 
    Increment, K, NodalForces, Residual, mesh, Eulerx, Eulerp, material,
    boundary_condition, AppliedDirichletInc, NeumannForces, TotalDisp):

    Tolerance = self.newton_raphson_tolerance
    LoadIncrement = self.number_of_load_increments
    Iter = 0

    # APPLY INCREMENTAL DIRICHLET PER LOAD STEP (THIS IS INCREMENTAL NOT ACCUMULATIVE)
    IncDirichlet = boundary_condition.UpdateFixDoFs(AppliedDirichletInc,
        K.shape[0],formulation.nvar)

    # UPDATE EULERIAN COORDINATE
    Eulerx += IncDirichlet[:,:formulation.ndim]
    Eulerp += IncDirichlet[:,-1]

    # Predictor
    if Increment == 0:
        # GET THE REDUCED SYSTEM OF EQUATIONS
        # K_b, F_b = boundary_condition.GetReducedMatrices(K,self.accumulated_load_factor*NeumannForces)[:2]
        K_b, F_b = boundary_condition.GetReducedMatrices(K,NeumannForces)[:2]
        # SOLVE THE SYSTEM
        sol = solver.Solve(K_b,F_b)
        # GET ITERATIVE SOLUTION
        dU = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar)
        # self.incremental_load_factor = 1./LoadIncrement
    else:
        dU = TotalDisp[:,:,Increment-1]*self.arc_length_scaling_factor
        self.incremental_load_factor *= self.arc_length_scaling_factor
        self.accumulated_load_factor += self.incremental_load_factor

    # UPDATE THE EULERIAN COMPONENTS
    Eulerx += dU[:,:formulation.ndim]
    Eulerp += dU[:,-1]


    while self.norm_residual > Tolerance or Iter==0:

        # GET THE REDUCED SYSTEM OF EQUATIONS
        K_b, F_b = boundary_condition.GetReducedMatrices(K,NeumannForces)[:2]
        # SOLVE THE SYSTEM
        sol = solver.Solve(K_b,F_b)
        # GET ITERATIVE SOLUTION
        dU1 = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar)

        # GET THE REDUCED SYSTEM OF EQUATIONS
        K_b, F_b = boundary_condition.GetReducedMatrices(K,Residual)[:2]
        # SOLVE THE SYSTEM
        sol = solver.Solve(K_b,-F_b)
        # GET ITERATIVE SOLUTION
        dU2 = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar)

        iterative_load_factor = -np.dot(dU.flatten(),dU2.flatten())/np.dot(dU.flatten(),dU1.flatten())
        ddU   = iterative_load_factor*dU1 + dU2
        # print(ddlam)
        # dU = dU2

        # UPDATE THE EULERIAN COMPONENTS
        self.incremental_load_factor += iterative_load_factor
        self.accumulated_load_factor += iterative_load_factor
        dU[:,:] += ddU[:,:]
        Eulerx += ddU[:,:formulation.ndim]
        Eulerp += ddU[:,-1]
        # Eulerx += dU[:,:formulation.ndim]
        # Eulerp += dU[:,-1]
        # print(self.accumulated_load_factor)

        # RE-ASSEMBLE - COMPUTE INTERNAL TRACTION FORCES
        K, TractionForces = Assemble(self, function_spaces[0], formulation, mesh, material,
            Eulerx,Eulerp)[:2]

        # FIND THE RESIDUAL
        # Residual[boundary_condition.columns_in] = TractionForces[boundary_condition.columns_in] -\
        #     NodalForces[boundary_condition.columns_in] - NeumannForces[boundary_condition.columns_in]*self.accumulated_load_factor
        Residual[boundary_condition.columns_in] = TractionForces[boundary_condition.columns_in] -\
            NeumannForces[boundary_condition.columns_in]*self.accumulated_load_factor

        # SAVE THE NORM
        self.rel_norm_residual = la.norm(Residual[boundary_condition.columns_in])
        if Iter==0:
            self.NormForces = la.norm(Residual[boundary_condition.columns_in])
        self.norm_residual = np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces) 

        # SAVE THE NORM 
        self.NRConvergence['Increment_'+str(Increment)] = np.append(self.NRConvergence['Increment_'+str(Increment)],\
            self.norm_residual)
        
        print("Iteration {} for increment {}.".format(Iter, Increment) +\
            " Residual (abs) {0:>16.7g}".format(self.rel_norm_residual), 
            "\t Residual (rel) {0:>16.7g}".format(self.norm_residual))

        if np.abs(self.rel_norm_residual) < Tolerance:
            break

        # UPDATE ITERATION NUMBER
        Iter +=1

        self.arc_length_scaling_factor = 0.5**(0.25*(Iter-5))

        if Iter==self.maximum_iteration_for_newton_raphson and formulation.fields == "electro_mechanics":
            # raise StopIteration("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
            warn("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
            self.newton_raphson_failed_to_converge = True
            break

        if Iter==self.maximum_iteration_for_newton_raphson:
            self.newton_raphson_failed_to_converge = True
            break
        if np.isnan(self.norm_residual) or self.norm_residual>1e06:
            self.newton_raphson_failed_to_converge = True
            break

        # USER DEFINED CRITERIA TO BREAK OUT OF NEWTON-RAPHSON
        if self.user_defined_break_func != None:
            if self.user_defined_break_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                break

        # USER DEFINED CRITERIA TO STOP NEWTON-RAPHSON AND THE WHOLE ANALYSIS
        if self.user_defined_stop_func != None:
            if self.user_defined_stop_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
                self.newton_raphson_failed_to_converge = True
                break


    return Eulerx, Eulerp, K, Residual



# def NewtonRaphsonArchLength(self, function_spaces, formulation, solver, 
#     Increment, K, NodalForces, Residual, mesh, Eulerx, Eulerp, material,
#     boundary_condition, AppliedDirichletInc, DeltaF, TotalDisp):

#     Tolerance = self.newton_raphson_tolerance
#     LoadIncrement = self.number_of_load_increments
#     LoadFactor = 1./LoadIncrement
#     accumulated_load_factor = Increment/LoadIncrement
#     Iter = 0
#     dL = 1.
#     psi = 1.
#     # NodalForces = DeltaF

#     Dlam = 0.
#     dU = np.zeros((mesh.points.shape[0],formulation.nvar))
#     dU_b = np.zeros((mesh.points.shape[0],formulation.nvar))

#     # SOLVE WITH INCREMENTAL LOAD
#     K_b, DF_b = boundary_condition.GetReducedMatrices(K,NodalForces)[:2]
#     dU_t = solver.Solve(K_b,DF_b)
#     dU_t = boundary_condition.UpdateFreeDoFs(dU_t,K.shape[0],formulation.nvar)
#     # print(NodalForces)

#     # dU = IncDirichlet
#     # GET TOTAL ITERATIVE SOLUTION
#     # dU = dU_actual + LoadFactor*dU_current

#     # GET ARC LENGTH QUADRATIC EQUATIONS COEFFICIENTS
#     # c1 = np.dot(dU.ravel(),dU.ravel()) + psi**2 * np.dot(DeltaF.ravel(),DeltaF.ravel())
#     # c2 = 2.*np.dot(DU.ravel()+dU_actual.ravel(),dU_current.ravel()) + 2.*psi**2 * LoadFactor * np.dot(DeltaF.ravel(),DeltaF.ravel())
#     # c3 = np.dot((DU+dU_actual).ravel(),(DU+dU_actual).ravel()) + psi**2 * LoadFactor**2 * np.dot(DeltaF.ravel(),DeltaF.ravel()) - dL**2
#     # coeffs = [c1,c2,c3]

#     # c1 = np.dot(dU_t.ravel(),dU_t.ravel()) + psi**2 * np.dot(NodalForces.ravel(),NodalForces.ravel())
#     # c2 = 2.*np.dot(dU.ravel()+dU_b.ravel(),dU_t.ravel()) + 2.*psi**2 * Dlam * np.dot(NodalForces.ravel(),NodalForces.ravel())
#     # c3 = np.dot((dU+dU_b).ravel(),(dU+dU_b).ravel()) + psi**2 * Dlam**2 * np.dot(NodalForces.ravel(),NodalForces.ravel()) - dL**2
#     # coeffs = [c1,c2,c3]

#     # # FIND THE NEW LOAD FACTOR
#     # dlams = np.roots(coeffs)
#     # dlam = np.real(dlams.max())
#     # # print(c1,c2,c3,dlams, dlam)

#     # # CORRECTOR
#     # dU_iter = dU_b + dlam*dU_t
#     # # print (dU_iter)
#     # # exit()


#     # APPLY INCREMENTAL DIRICHLET PER LOAD STEP (THIS IS INCREMENTAL NOT ACCUMULATIVE)
#     IncDirichlet = boundary_condition.UpdateFixDoFs(AppliedDirichletInc,
#         K.shape[0],formulation.nvar)

#     # UPDATE EULERIAN COORDINATE
#     Eulerx += IncDirichlet[:,:formulation.ndim]
#     Eulerp += IncDirichlet[:,-1]
#     # Eulerx += IncDirichlet[:,:formulation.ndim] + dU_iter[:,:formulation.ndim]
#     # Eulerp += IncDirichlet[:,-1] + dU_iter[:,-1]

#     # accumulated_load_factor += dlam
#     # if Increment>0:
#     #     DU = TotalDisp[:,:,Increment] - TotalDisp[:,:,Increment-1]
#     # else:
#     #     DU = np.zeros((mesh.points.shape[0],formulation.nvar))

#     # DU = np.zeros((mesh.points.shape[0],formulation.nvar))


#     while self.norm_residual > Tolerance or Iter==0:
#         # GET THE REDUCED SYSTEM OF EQUATIONS
#         K_b, F_b = boundary_condition.GetReducedMatrices(K,Residual)[:2]
#         # SOLVE THE SYSTEM
#         sol = solver.Solve(K_b,-F_b)
#         # GET ITERATIVE SOLUTION
#         # dU_b = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar) 
#         dU = boundary_condition.UpdateFreeDoFs(sol,K.shape[0],formulation.nvar)


#         # print(dlams)
#         # exit()
#         # LoadFactor += np.real(np.max(dlams))
#         # print(LoadFactor)

#         c1 = np.dot(dU_t.ravel(),dU_t.ravel()) + psi**2 * np.dot(NodalForces.ravel(),NodalForces.ravel())
#         c2 = 2.*np.dot(dU.ravel()+dU_b.ravel(),dU_t.ravel()) + 2.*psi**2 * Dlam * np.dot(NodalForces.ravel(),NodalForces.ravel())
#         c3 = np.dot((dU+dU_b).ravel(),(dU+dU_b).ravel()) + psi**2 * Dlam**2 * np.dot(NodalForces.ravel(),NodalForces.ravel()) - dL**2
#         coeffs = [c1,c2,c3]

#         # FIND THE NEW LOAD FACTOR
#         dlams = np.roots(coeffs)
#         dlam = np.real(dlams.max())
#         print(dlam)

#         # CORRECTOR
#         dU_iter = dU_b + dlam*dU_t
#         accumulated_load_factor += dlam


#         # UPDATE THE EULERIAN COMPONENTS
#         Eulerx += dU[:,:formulation.ndim]
#         Eulerp += dU[:,-1]
#         # Eulerx += dU_iter[:,:formulation.ndim]
#         # Eulerp += dU_iter[:,-1]

#         # RE-ASSEMBLE - COMPUTE INTERNAL TRACTION FORCES
#         K, TractionForces = Assemble(self, function_spaces[0], formulation, mesh, material,
#             Eulerx,Eulerp)[:2]

#         # FIND THE RESIDUAL
#         Residual[boundary_condition.columns_in] = TractionForces[boundary_condition.columns_in] -\
#             NodalForces[boundary_condition.columns_in]

#         # SAVE THE NORM
#         self.rel_norm_residual = la.norm(Residual[boundary_condition.columns_in])
#         if Iter==0:
#             self.NormForces = la.norm(Residual[boundary_condition.columns_in])
#         self.norm_residual = np.abs(la.norm(Residual[boundary_condition.columns_in])/self.NormForces) 

#         # SAVE THE NORM 
#         self.NRConvergence['Increment_'+str(Increment)] = np.append(self.NRConvergence['Increment_'+str(Increment)],\
#             self.norm_residual)
        
#         print("Iteration {} for increment {}.".format(Iter, Increment) +\
#             " Residual (abs) {0:>16.7g}".format(self.rel_norm_residual), 
#             "\t Residual (rel) {0:>16.7g}".format(self.norm_residual))

#         if np.abs(self.rel_norm_residual) < Tolerance:
#             break

#         # UPDATE ITERATION NUMBER
#         Iter +=1

#         if Iter==self.maximum_iteration_for_newton_raphson and formulation.fields == "electro_mechanics":
#             # raise StopIteration("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
#             warn("\n\nNewton Raphson did not converge! Maximum number of iterations reached.")
#             self.newton_raphson_failed_to_converge = True
#             break

#         if Iter==self.maximum_iteration_for_newton_raphson:
#             self.newton_raphson_failed_to_converge = True
#             break
#         if np.isnan(self.norm_residual) or self.norm_residual>1e06:
#             self.newton_raphson_failed_to_converge = True
#             break

#         # USER DEFINED CRITERIA TO BREAK OUT OF NEWTON-RAPHSON
#         if self.user_defined_break_func != None:
#             if self.user_defined_break_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
#                 break

#         # USER DEFINED CRITERIA TO STOP NEWTON-RAPHSON AND THE WHOLE ANALYSIS
#         if self.user_defined_stop_func != None:
#             if self.user_defined_stop_func(Increment,Iter,self.norm_residual,self.rel_norm_residual, Tolerance):
#                 self.newton_raphson_failed_to_converge = True
#                 break


#     return Eulerx, Eulerp, K, Residual